from abc import ABC
from typing import Literal, Optional

from torchtyping import TensorType

from gfn.containers import States
from gfn.envs import Env
from gfn.envs.preprocessors.base import EnumPreprocessor
from gfn.modules import GFNModule, NeuralNet, Tabular, Uniform, ZeroGFNModule

# Typing
OutputTensor = TensorType["batch_shape", "output_dim", float]


class FunctionEstimator(ABC):
    """Base class for function estimators."""

    def __init__(
        self,
        env: Env,
        module: Optional[GFNModule] = None,
        output_dim: Optional[int] = None,
        module_name: Optional[
            Literal["NeuralNet", "Uniform", "Tabular", "Zero"]
        ] = None,
        **nn_kwargs,
    ) -> None:
        """Either module or (module_name, output_dim) must be provided.

        Args:
            env (Env): the environment.
            module (Optional[GFNModule], optional): The module to use. Defaults to None.
            output_dim (Optional[int], optional): Used only if module is None. Defines the output dimension of the module. Defaults to None.
            module_name (Optional[Literal[NeuralNet, Uniform, Tabular, Zero]], optional): Used only if module is None. What module to use. Defaults to None.
            **nn_kwargs: Keyword arguments to pass to the module, if module_name is NeuralNet.
        """
        self.env = env
        if module is None:
            assert module_name is not None and output_dim is not None
            if module_name == "NeuralNet":
                assert len(env.preprocessor.output_shape) == 1
                input_dim = env.preprocessor.output_shape[0]
                module = NeuralNet(
                    input_dim=input_dim,
                    output_dim=output_dim,
                    **nn_kwargs,
                )
            elif module_name == "Uniform":
                module = Uniform(output_dim=output_dim)
            elif module_name == "Zero":
                module = ZeroGFNModule(output_dim=output_dim)
            elif module_name == "Tabular":
                module = Tabular(
                    n_states=env.n_states,
                    output_dim=output_dim,
                )
            else:
                raise ValueError(f"Unknown module_name {module_name}")
        self.module = module
        if isinstance(self.module, Tabular):
            self.preprocessor = EnumPreprocessor(env.get_states_indices)
        else:
            self.preprocessor = env.preprocessor

    def __call__(self, states: States) -> OutputTensor:
        return self.module(self.preprocessor(states))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.module})"

    def named_parameters(self) -> dict:
        return dict(self.module.named_parameters())

    def load_state_dict(self, state_dict: dict):
        self.module.load_state_dict(state_dict)


class LogEdgeFlowEstimator(FunctionEstimator):
    r"""Container for estimators $(s \rightarrow s') \mapsto \log F(s \rightarrow s')$.
    The way it's coded is a function $s \mapsto (\log F(s \rightarrow (s + a)))_{a \in \mathbb{A}}$,
    where $s+a$ is the state obtained by performing action $a$ in state $s$."""

    def __init__(
        self,
        env: Env,
        module: Optional[GFNModule] = None,
        module_name: Optional[
            Literal["NeuralNet", "Uniform", "Tabular", "Zero"]
        ] = None,
        **nn_kwargs,
    ) -> None:
        if module is not None:
            assert module.output_dim == env.n_actions
        super().__init__(
            env,
            module=module,
            output_dim=env.n_actions,
            module_name=module_name,
            **nn_kwargs,
        )


class LogStateFlowEstimator(FunctionEstimator):
    r"""Container for estimators $s \mapsto \log F(s)$."""

    def __init__(
        self,
        env: Env,
        module: Optional[GFNModule] = None,
        module_name: Optional[
            Literal["NeuralNet", "Uniform", "Tabular", "Zero"]
        ] = None,
        forward_looking=False,
        **nn_kwargs,
    ):
        if module is not None:
            assert module.output_dim == 1
        super().__init__(
            env, module=module, output_dim=1, module_name=module_name, **nn_kwargs
        )
        self.forward_looking = forward_looking

    def __call__(self, states: States) -> OutputTensor:
        out = super().__call__(states)
        if self.forward_looking:
            log_rewards = self.env.log_reward(states).unsqueeze(-1)
            out = out + log_rewards
        return out


class LogitPFEstimator(FunctionEstimator):
    r"""Container for estimators $s \mapsto (u(s + a \mid s))_{a \in \mathbb{A}}$,
    such that $P_F(s + a \mid s) = \frac{e^{u(s + a \mid s)}}{\sum_{a' \in \mathbb{A}} e^{u(s + a' \mid s)}}$."""

    def __init__(
        self,
        env: Env,
        module: Optional[GFNModule] = None,
        module_name: Optional[
            Literal["NeuralNet", "Uniform", "Tabular", "Zero"]
        ] = None,
        **nn_kwargs,
    ):
        if module is not None:
            assert module.output_dim == env.n_actions
        super().__init__(
            env,
            module=module,
            output_dim=env.n_actions,
            module_name=module_name,
            **nn_kwargs,
        )


class LogitPBEstimator(FunctionEstimator):
    r"""Container for estimators $s \mapsto (u(s' - a \mid s'))_{a \in \mathbb{A}}$,
    such that $P_B(s' - a \mid s') = \frac{e^{u(s' - a \mid s')}}{\sum_{a' \in \mathbb{A}} e^{u(s' - a' \mid s')}}$."""

    def __init__(
        self,
        env: Env,
        module: Optional[GFNModule] = None,
        module_name: Optional[
            Literal["NeuralNet", "Uniform", "Tabular", "Zero"]
        ] = None,
        **nn_kwargs,
    ):
        if module is not None:
            assert module.output_dim == env.n_actions - 1
        super().__init__(
            env,
            module=module,
            output_dim=env.n_actions - 1,
            module_name=module_name,
            **nn_kwargs,
        )


class LogZEstimator:
    r"""Container for the estimator $\log Z$."""

    def __init__(self, tensor: TensorType[0, float]) -> None:
        self.tensor = tensor
        assert self.tensor.shape == ()
        self.tensor.requires_grad = True

    def __repr__(self) -> str:
        return str(self.tensor.item())

    def named_parameters(self) -> dict:
        return {"logZ": self.tensor}

    def load_state_dict(self, state_dict: dict):
        self.tensor = state_dict["logZ"]



from typing import List, Optional

import torch
from torchtyping import TensorType

from gfn.containers import States, Trajectories
from gfn.envs import Env
from gfn.samplers.actions_samplers import ActionsSampler, BackwardActionsSampler

# Typing
StatesTensor = TensorType["n_trajectories", "state_shape", torch.float]
ActionsTensor = TensorType["n_trajectories", torch.long]
LogProbsTensor = TensorType["n_trajectories", torch.float]
DonesTensor = TensorType["n_trajectories", torch.bool]


class TrajectoriesSampler:
    def __init__(
        self,
        env: Env,
        actions_sampler: ActionsSampler,
    ):
        """Sample complete trajectories, or completes trajectories from a given batch states, using actions_sampler.

        Args:
            env (Env): Environment to sample trajectories from.
            actions_sampler (ActionsSampler): Sampler of actions.
        """
        self.env = env
        self.actions_sampler = actions_sampler
        self.is_backward = isinstance(actions_sampler, BackwardActionsSampler)

    def sample_trajectories(
        self,
        states: Optional[States] = None,
        n_trajectories: Optional[int] = None,
    ) -> Trajectories:
        if states is None:
            assert (
                n_trajectories is not None
            ), "Either states or n_trajectories should be specified"
            states = self.env.reset(batch_shape=(n_trajectories,))
        else:
            assert (
                len(states.batch_shape) == 1
            ), "States should be a linear batch of states"
            n_trajectories = states.batch_shape[0]

        device = states.states_tensor.device

        dones = states.is_initial_state if self.is_backward else states.is_sink_state

        trajectories_states: List[StatesTensor] = [states.states_tensor]
        trajectories_actions: List[ActionsTensor] = []
        trajectories_logprobs: List[LogProbsTensor] = []
        trajectories_dones = torch.zeros(
            n_trajectories, dtype=torch.long, device=device
        )
        trajectories_log_rewards = torch.zeros(
            n_trajectories, dtype=torch.float, device=device
        )

        step = 0

        while not all(dones):
            actions = torch.full(
                (n_trajectories,),
                fill_value=-1,
                dtype=torch.long,
                device=device,
            )
            log_probs = torch.full(
                (n_trajectories,), fill_value=0, dtype=torch.float, device=device
            )
            actions_log_probs, valid_actions = self.actions_sampler.sample(
                states[~dones]
            )
            actions[~dones] = valid_actions
            log_probs[~dones] = actions_log_probs
            trajectories_actions += [actions]
            trajectories_logprobs += [log_probs]

            if self.is_backward:
                new_states = self.env.backward_step(states, actions)
            else:
                new_states = self.env.step(states, actions)
            sink_states_mask = new_states.is_sink_state

            step += 1

            new_dones = (
                new_states.is_initial_state if self.is_backward else sink_states_mask
            ) & ~dones
            trajectories_dones[new_dones & ~dones] = step
            try:
                trajectories_log_rewards[new_dones & ~dones] = self.env.log_reward(
                    states[new_dones & ~dones]
                )
            except NotImplementedError:
                # print(states[new_dones & ~dones])
                # print(torch.log(self.env.reward(states[new_dones & ~dones])))
                trajectories_log_rewards[new_dones & ~dones] = torch.log(
                    self.env.reward(states[new_dones & ~dones])
                )
            states = new_states
            dones = dones | new_dones

            trajectories_states += [states.states_tensor]

        trajectories_states = torch.stack(trajectories_states, dim=0)
        trajectories_states = self.env.States(states_tensor=trajectories_states)
        trajectories_actions = torch.stack(trajectories_actions, dim=0)
        trajectories_logprobs = torch.stack(trajectories_logprobs, dim=0)

        trajectories = Trajectories(
            env=self.env,
            states=trajectories_states,
            actions=trajectories_actions,
            when_is_done=trajectories_dones,
            is_backward=self.is_backward,
            log_rewards=trajectories_log_rewards,
            log_probs=trajectories_logprobs,
        )

        return trajectories

    def sample(self, n_trajectories: int) -> Trajectories:
        return self.sample_trajectories(n_trajectories=n_trajectories)



from abc import ABC, abstractmethod
from typing import Callable, Tuple

from torchtyping import TensorType

from gfn.containers import States

# Typing
OutputTensor = TensorType["batch_shape", "dim_in"]


class Preprocessor(ABC):
    """
    Base class for Preprocessors. The goal is to transform tensors representing raw states
    to tensors that can be used as input to neural networks.
    """

    name: str = "Preprocessor"

    def __init__(self, output_shape: Tuple[int]) -> None:
        self.output_shape = output_shape

    @abstractmethod
    def preprocess(self, states: States) -> OutputTensor:
        pass

    def __call__(self, states: States) -> OutputTensor:
        return self.preprocess(states)

    def __repr__(self):
        return f"{self.name}, output_shape={self.output_shape}"


class IdentityPreprocessor(Preprocessor):
    """Simple preprocessor applicable to environments with uni-dimensional states.
    This is the default preprocessor used."""

    name = "IdentityPreprocessor"

    def preprocess(self, states: States) -> OutputTensor:
        return states.states_tensor.float()


class EnumPreprocessor(Preprocessor):
    "Preprocessor applicable to environments with discrete states."
    name = "EnumPreprocessor"

    def __init__(self, get_states_indices: Callable[[States], OutputTensor]) -> None:
        """Preprocessor for environments with enumerable states (finite number of states).
        Each state is represented by a unique integer (>= 0) index.

        Args:
            get_states_indices (Callable[[States], OutputTensor]): function that returns the unique indices of the states.
        """
        super().__init__(output_shape=(1,))
        self.get_states_indices = get_states_indices

    def preprocess(self, states):
        return self.get_states_indices(states).long().unsqueeze(-1)



from __future__ import annotations

from abc import ABC
from math import prod
from typing import ClassVar, Sequence, cast

import torch
from torchtyping import TensorType

from gfn.containers.base import Container

# Typing
ForwardMasksTensor = TensorType["batch_shape", "n_actions", torch.bool]
BackwardMasksTensor = TensorType["batch_shape", "n_actions - 1", torch.bool]
StatesTensor = TensorType["batch_shape", "state_shape", torch.float]
DonesTensor = TensorType["batch_shape", torch.bool]
RewardsTensor = TensorType["batch_shape", torch.float]
OneStateTensor = TensorType["state_shape", torch.float]


def correct_cast(
    forward_masks: ForwardMasksTensor | None,
    backward_masks: BackwardMasksTensor | None,
) -> tuple[ForwardMasksTensor, BackwardMasksTensor]:
    """
    Casts the given masks to the correct type, if they are not None.
    This function is to help with type checking only.
    """
    forward_masks = cast(ForwardMasksTensor, forward_masks)
    backward_masks = cast(BackwardMasksTensor, backward_masks)
    return forward_masks, backward_masks


class States(Container, ABC):
    """Base class for states, seen as nodes of the DAG.
    For each environment, a States subclass is needed. A `States` object
    is a collection of multiple states (nodes of the DAG). A tensor representation
    of the states is required for batching. If a state is represented with a tensor
    of shape (*state_shape), a batch of states is represented with a States object,
    with the attribute `states_tensor` of shape (*batch_shape, *state_shape). Other
    representations are possible (e.g. state as string, as numpy array, as graph, etc...),
    but these representations should not be batched.

    If the environment's action space is discrete, then each States object is also endowed
    with a `forward_masks` and `backward_masks` boolean attributes representing which actions
    are allowed at each state.

    A `batch_shape` attribute is also required, to keep track of the batch dimension.
    A trajectory can be represented by a States object with batch_shape = (n_states,).
    Multiple trajectories can be represented by a States object with batch_shape = (n_states, n_trajectories).

    Because multiple trajectories can have different lengths, batching requires appending a dummy tensor
    to trajectories that are shorter than the longest trajectory. The dummy state is the `s_f`
    attribute of the environment (e.g. [-1, ..., -1], or [-inf, ..., -inf], etc...). Which is never processed,
    and is used to pad the batch of states only.
    """

    state_shape: ClassVar[tuple[int, ...]]  # Shape of one state
    s0: ClassVar[OneStateTensor]  # Source state of the DAG
    sf: ClassVar[OneStateTensor]  # Dummy state, used to pad a batch of states

    def __init__(
        self,
        states_tensor: StatesTensor,
        forward_masks: ForwardMasksTensor | None = None,
        backward_masks: BackwardMasksTensor | None = None,
    ):
        self.states_tensor = states_tensor
        self.batch_shape = tuple(self.states_tensor.shape)[: -len(self.state_shape)]
        if forward_masks is None and backward_masks is None:
            try:
                self.forward_masks, self.backward_masks = self.make_masks()
                self.update_masks()
            except NotImplementedError:
                pass
        else:
            self.forward_masks = forward_masks
            self.backward_masks = backward_masks

        self._log_rewards = (
            None  # Useful attribute if we want to store the log-reward of the states
        )

    @classmethod
    def from_batch_shape(cls, batch_shape: tuple[int], random: bool = False) -> States:
        """Create a States object with the given batch shape, all initialized to s_0.
        If random is True, the states are initialized randomly. This requires that
        the environment implements the `make_random_states_tensor` class method.
        """
        if random:
            states_tensor = cls.make_random_states_tensor(batch_shape)
        else:
            states_tensor = cls.make_initial_states_tensor(batch_shape)
        return cls(states_tensor)

    @classmethod
    def make_initial_states_tensor(cls, batch_shape: tuple[int]) -> StatesTensor:
        state_ndim = len(cls.state_shape)
        assert cls.s0 is not None and state_ndim is not None
        return cls.s0.repeat(*batch_shape, *((1,) * state_ndim))

    @classmethod
    def make_random_states_tensor(cls, batch_shape: tuple[int]) -> StatesTensor:
        raise NotImplementedError(
            "The environment does not support initialization of random states."
        )

    def make_masks(self) -> tuple[ForwardMasksTensor, BackwardMasksTensor]:
        """Create the forward and backward masks for the states.
        This method is called only if the masks are not provided at initialization.
        """
        return NotImplementedError(
            "make_masks method not implemented. Your environment must implement it if discrete"
        )

    def update_masks(self) -> None:
        """Update the masks, if necessary.
        This method should be called after each action is taken.
        """
        return NotImplementedError(
            "update_masks method not implemented. Your environment must implement it if discrete"
        )

    def __len__(self):
        return prod(self.batch_shape)

    def __repr__(self):
        return f"{self.__class__.__name__} object of batch shape {self.batch_shape} and state shape {self.state_shape}"

    @property
    def device(self) -> torch.device:
        return self.states_tensor.device

    def __getitem__(self, index: int | Sequence[int] | Sequence[bool]) -> States:
        """Access particular states of the batch."""
        # TODO: add more tests for this method
        states = self.states_tensor[index]
        if self.forward_masks is None and self.backward_masks is None:
            return self.__class__(states)
        else:
            self.forward_masks, self.backward_masks = correct_cast(
                self.forward_masks, self.backward_masks
            )
            forward_masks = self.forward_masks[index]
            backward_masks = self.backward_masks[index]
            return self.__class__(
                states, forward_masks=forward_masks, backward_masks=backward_masks
            )

    def flatten(self) -> States:
        """Flatten the batch dimension of the states.
        This is useful for example when extracting individual states from trajectories.
        """
        states = self.states_tensor.view(-1, *self.state_shape)
        if self.forward_masks is None and self.backward_masks is None:
            return self.__class__(states)
        else:
            self.forward_masks, self.backward_masks = correct_cast(
                self.forward_masks, self.backward_masks
            )
            forward_masks = self.forward_masks.view(-1, self.forward_masks.shape[-1])
            backward_masks = self.backward_masks.view(-1, self.backward_masks.shape[-1])
            return self.__class__(
                states, forward_masks=forward_masks, backward_masks=backward_masks
            )

    def extend(self, other: States) -> None:
        """Collates to another States object of the same batch shape, which should be 1 or 2.

        Args:
            other (States): Batch of states to collate to.

        Raises:
            ValueError: if self.batch_shape != other.batch_shape or if self.batch_shape != (1,) or (2,)
        """
        other_batch_shape = other.batch_shape
        if len(other_batch_shape) == len(self.batch_shape) == 1:
            # This corresponds to adding a state to a trajectory
            self.batch_shape = (self.batch_shape[0] + other_batch_shape[0],)
            self.states_tensor = torch.cat(
                (self.states_tensor, other.states_tensor), dim=0
            )

        elif len(other_batch_shape) == len(self.batch_shape) == 2:
            # This corresponds to adding a trajectory to a batch of trajectories
            self.extend_with_sf(
                required_first_dim=max(self.batch_shape[0], other_batch_shape[0])
            )
            other.extend_with_sf(
                required_first_dim=max(self.batch_shape[0], other_batch_shape[0])
            )
            self.batch_shape = (
                self.batch_shape[0],
                self.batch_shape[1] + other_batch_shape[1],
            )
            self.states_tensor = torch.cat(
                (self.states_tensor, other.states_tensor), dim=1
            )
        else:
            raise ValueError(
                f"extend is not implemented for batch shapes {self.batch_shape} and {other_batch_shape}"
            )
        if self.forward_masks is not None and self.backward_masks is not None:
            self.forward_masks, self.backward_masks = correct_cast(
                self.forward_masks, self.backward_masks
            )
            other.forward_masks, other.backward_masks = correct_cast(
                other.forward_masks, other.backward_masks
            )
            self.forward_masks = torch.cat(
                (self.forward_masks, other.forward_masks), dim=len(self.batch_shape) - 1
            )
            self.backward_masks = torch.cat(
                (self.backward_masks, other.backward_masks),
                dim=len(self.batch_shape) - 1,
            )

    def extend_with_sf(self, required_first_dim: int) -> None:
        """Takes a two-dimensional batch of states (i.e. of batch_shape (a, b)),
        and extends it to a States object of batch_shape (required_first_dim, b),
        by adding the required number of `s_f` tensors. This is useful to extend trajectories
        of different lengths."""
        if len(self.batch_shape) == 2:
            if self.batch_shape[0] >= required_first_dim:
                return
            self.states_tensor = torch.cat(
                (
                    self.states_tensor,
                    self.__class__.sf.repeat(
                        required_first_dim - self.batch_shape[0], self.batch_shape[1], 1
                    ),
                ),
                dim=0,
            )
            if self.forward_masks is not None and self.backward_masks is not None:
                self.forward_masks, self.backward_masks = correct_cast(
                    self.forward_masks, self.backward_masks
                )
                self.forward_masks = torch.cat(
                    (
                        self.forward_masks,
                        torch.ones(
                            required_first_dim - self.batch_shape[0],
                            *self.forward_masks.shape[1:],
                            dtype=torch.bool,
                            device=self.device,
                        ),
                    ),
                    dim=0,
                )
                self.backward_masks = torch.cat(
                    (
                        self.backward_masks,
                        torch.ones(
                            required_first_dim - self.batch_shape[0],
                            *self.backward_masks.shape[1:],
                            dtype=torch.bool,
                            device=self.device,
                        ),
                    ),
                    dim=0,
                )
            self.batch_shape = (required_first_dim, self.batch_shape[1])
        else:
            raise ValueError(
                f"extend_with_sf is not implemented for batch shapes {self.batch_shape}"
            )

    def compare(self, other: StatesTensor) -> DonesTensor:
        """Given a tensor of states, returns a tensor of booleans indicating whether the states
        are equal to the states in self.

        Args:
            other (StatesTensor): Tensor of states to compare to.

        Returns:
            DonesTensor: Tensor of booleans indicating whether the states are equal to the states in self.
        """
        out = self.states_tensor == other
        state_ndim = len(self.__class__.state_shape)
        for _ in range(state_ndim):
            out = out.all(dim=-1)
        return out

    @property
    def is_initial_state(self) -> DonesTensor:
        r"""Return a boolean tensor of shape=(*batch_shape,),
        where True means that the state is $s_0$ of the DAG.
        """
        source_states_tensor = self.__class__.s0.repeat(
            *self.batch_shape, *((1,) * len(self.__class__.state_shape))
        )
        return self.compare(source_states_tensor)

    @property
    def is_sink_state(self) -> DonesTensor:
        r"""Return a boolean tensor of shape=(*batch_shape,),
        where True means that the state is $s_f$ of the DAG.
        """
        sink_states = self.__class__.sf.repeat(
            *self.batch_shape, *((1,) * len(self.__class__.state_shape))
        )
        return self.compare(sink_states)

    @property
    def log_rewards(self) -> RewardsTensor:
        return self._log_rewards

    @log_rewards.setter
    def log_rewards(self, log_rewards: RewardsTensor) -> None:
        self._log_rewards = log_rewards



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Literal

from gfn.containers.states import States
from gfn.containers.trajectories import Trajectories
from gfn.containers.transitions import Transitions
from gfn.losses.base import (
    EdgeDecomposableLoss,
    Loss,
    StateDecomposableLoss,
    TrajectoryDecomposableLoss,
)

if TYPE_CHECKING:
    from gfn.envs import Env


class ReplayBuffer:
    def __init__(
        self,
        env: Env,
        loss_fn: Loss | None = None,
        objects_type: Literal["transitions", "trajectories", "states"] | None = None,
        capacity: int = 1000,
    ):
        self.env = env
        self.capacity = capacity
        self.terminating_states = None
        if objects_type == "trajectories" or isinstance(
            loss_fn, TrajectoryDecomposableLoss
        ):
            self.training_objects = Trajectories(env)
            self.objects_type = "trajectories"
        elif objects_type == "transitions" or isinstance(loss_fn, EdgeDecomposableLoss):
            self.training_objects = Transitions(env)
            self.objects_type = "transitions"
        elif objects_type == "states" or isinstance(loss_fn, StateDecomposableLoss):
            self.training_objects = env.States.from_batch_shape((0,))
            self.terminating_states = env.States.from_batch_shape((0,))
            self.objects_type = "states"
        else:
            raise ValueError(
                f"Unknown objects_type: {objects_type} and loss_fn: {loss_fn}"
            )

        self._is_full = False
        self._index = 0

    def __repr__(self):
        return f"ReplayBuffer(capacity={self.capacity}, containing {len(self)} {self.objects_type})"

    def __len__(self):
        return self.capacity if self._is_full else self._index

    def add(self, training_objects: Transitions | Trajectories | tuple[States]):
        terminating_states = None
        if isinstance(training_objects, tuple):
            assert self.objects_type == "states" and self.terminating_states is not None
            training_objects, terminating_states = training_objects

        to_add = len(training_objects)

        self._is_full |= self._index + to_add >= self.capacity
        self._index = (self._index + to_add) % self.capacity

        self.training_objects.extend(training_objects)
        self.training_objects = self.training_objects[-self.capacity :]

        if self.terminating_states is not None:
            assert terminating_states is not None
            self.terminating_states.extend(terminating_states)
            self.terminating_states = self.terminating_states[-self.capacity :]

    def sample(self, n_trajectories: int) -> Transitions | Trajectories | tuple[States]:
        if self.terminating_states is not None:
            return (
                self.training_objects.sample(n_trajectories),
                self.terminating_states.sample(n_trajectories),
            )
        return self.training_objects.sample(n_trajectories)

    def save(self, directory: str):
        self.training_objects.save(os.path.join(directory, "training_objects"))
        if self.terminating_states is not None:
            self.terminating_states.save(os.path.join(directory, "terminating_states"))

    def load(self, directory: str):
        self.training_objects.load(os.path.join(directory, "training_objects"))
        self._index = len(self.training_objects)
        if self.terminating_states is not None:
            self.terminating_states.load(os.path.join(directory, "terminating_states"))



from typing import Dict, Optional

import torch

from gfn.containers import States, Trajectories, Transitions
from gfn.distributions import EmpiricalTerminatingStatesDistribution
from gfn.envs import Env
from gfn.losses import (
    EdgeDecomposableLoss,
    Loss,
    Parametrization,
    StateDecomposableLoss,
    TBParametrization,
    TrajectoryDecomposableLoss,
)


def trajectories_to_training_samples(
    trajectories: Trajectories, loss_fn: Loss
) -> States | Transitions | Trajectories:
    """Converts a Trajectories container to a States, Transitions or Trajectories container,
    depending on the loss.
    """
    if isinstance(loss_fn, StateDecomposableLoss):
        # return trajectories.to_states()
        return trajectories.to_non_initial_intermediary_and_terminating_states()
    elif isinstance(loss_fn, TrajectoryDecomposableLoss):
        return trajectories
    elif isinstance(loss_fn, EdgeDecomposableLoss):
        return trajectories.to_transitions()
    else:
        raise ValueError(f"Loss {loss_fn} is not supported.")


def validate(
    env: Env,
    parametrization: Parametrization,
    n_validation_samples: int = 1000,
    visited_terminating_states: Optional[States] = None,
) -> Dict[str, float]:
    """Evaluates the current parametrization on the given environment.
    This is for environments with known target reward. The validation is done by computing the l1 distance between the
    learned empirical and the target distributions.

    Args:
        env: The environment to evaluate the parametrization on.
        parametrization: The parametrization to evaluate.
        n_validation_samples: The number of samples to use to evaluate the pmf.
        visited_terminating_states: The terminating states visited during training. If given, the pmf is obtained from
            these last n_validation_samples states. Otherwise, n_validation_samples are resampled for evaluation.

    Returns:
        Dict[str, float]: A dictionary containing the l1 validation metric. If the parametrization is a TBParametrization,
        i.e. contains LogZ, then the (absolute) difference between the learned and the target LogZ is also returned in the
        dictionary.
    """

    true_logZ = env.log_partition
    true_dist_pmf = env.true_dist_pmf
    if isinstance(true_dist_pmf, torch.Tensor):
        true_dist_pmf = true_dist_pmf.cpu()
    else:
        # The environment does not implement a true_dist_pmf property, nor a log_partition property
        # We cannot validate the parametrization
        return {}

    logZ = None
    if isinstance(parametrization, TBParametrization):
        logZ = parametrization.logZ.tensor.item()
    if visited_terminating_states is None:
        final_states_dist = parametrization.P_T(env, n_validation_samples)
    else:
        final_states_dist = EmpiricalTerminatingStatesDistribution(
            env, visited_terminating_states[-n_validation_samples:]
        )
        n_visited_states = visited_terminating_states.batch_shape[0]
        n_validation_samples = min(n_visited_states, n_validation_samples)

    final_states_dist_pmf = final_states_dist.pmf()
    l1_dist = (final_states_dist_pmf - true_dist_pmf).abs().mean().item()
    validation_info = {"l1_dist": l1_dist}
    if logZ is not None:
        validation_info["logZ_diff"] = abs(logZ - true_logZ)
    return validation_info



from typing import Callable

import torch
from einops import rearrange
from torch.nn.functional import one_hot
from torchtyping import TensorType

from gfn.containers.states import States
from gfn.envs.preprocessors.base import Preprocessor

# Typing
OutputTensor = TensorType["batch_shape", "dim_in"]


class OneHotPreprocessor(Preprocessor):
    name = "one_hot"

    def __init__(
        self,
        n_states: int,
        get_states_indices: Callable[[States], OutputTensor],
    ) -> None:
        """One Hot Preprocessor for environments with enumerable states (finite number of states).

        Args:
            n_states (int): The total number of states in the environment (not including s_f).
            get_states_indices (Callable[[States], OutputTensor]): function that returns the unique indices of the states.
        """
        super().__init__(output_shape=(n_states,))
        self.get_states_indices = get_states_indices
        self.output_dim = n_states

    def preprocess(self, states):
        state_indices = self.get_states_indices(states)
        return one_hot(state_indices, self.output_dim).float()


class KHotPreprocessor(Preprocessor):
    name = "k_hot"

    def __init__(
        self,
        height: int,
        ndim: int,
        get_states_indices: Callable[[States], OutputTensor],
    ) -> None:
        """K Hot Preprocessor for environments with enumerable states (finite number of states) with a grid structure.

        Args:
            height (int): number of unique values per dimension.
            ndim (int): number of dimensions.
            get_states_indices (Callable[[States], OutputTensor]): function that returns the unique indices of the states.
        """
        super().__init__(output_shape=(height * ndim,))
        self.height = height
        self.ndim = ndim
        self.get_states_indices = get_states_indices

    def preprocess(self, states):
        states_tensor = states.states_tensor
        assert (
            states_tensor.dtype == torch.long
        ), "K Hot preprocessing only works for integer states"
        states_tensor = states_tensor.long()
        hot = one_hot(states_tensor, self.height).float()
        hot = rearrange(hot, "... a b -> ... (a b)")
        return hot



from dataclasses import dataclass
from typing import Tuple

import torch
from torchtyping import TensorType

from gfn.containers.states import States, correct_cast
from gfn.distributions import EmpiricalTrajectoryDistribution, TrajectoryDistribution
from gfn.envs import Env
from gfn.estimators import LogEdgeFlowEstimator
from gfn.losses.base import Parametrization, StateDecomposableLoss
from gfn.samplers import DiscreteActionsSampler, TrajectoriesSampler

# Typing
ScoresTensor = TensorType["n_states", float]
LossTensor = TensorType[0, float]


@dataclass
class FMParametrization(Parametrization):
    r"""
    $\mathcal{O}_{edge}$ is the set of functions from the non-terminating edges
    to $\mathbb{R}^+$. Which is equivalent to the set of functions from the internal nodes
    (i.e. without $s_f$) to $(\mathbb{R})^{n_actions}$, without the exit action (No need for
    positivity if we parametrize log-flows).
    """
    logF: LogEdgeFlowEstimator

    def Pi(
        self, env: Env, n_samples: int = 1000, **actions_sampler_kwargs
    ) -> TrajectoryDistribution:
        actions_sampler = DiscreteActionsSampler(self.logF, **actions_sampler_kwargs)
        trajectories_sampler = TrajectoriesSampler(env, actions_sampler)
        trajectories = trajectories_sampler.sample_trajectories(
            n_trajectories=n_samples
        )
        return EmpiricalTrajectoryDistribution(trajectories)


class FlowMatching(StateDecomposableLoss):
    def __init__(self, parametrization: FMParametrization, alpha=1.0) -> None:
        "alpha is the weight of the reward matching loss"
        self.parametrization = parametrization
        self.env = parametrization.logF.env
        self.alpha = alpha

    def flow_matching_loss(self, states: States) -> ScoresTensor:
        """
        Compute the FM for the given states, defined as the log-sum incoming flows minus log-sum outgoing flows.
        The states should not include s0. The batch shape should be (n_states,).

        As of now, only discrete environments are handled.
        """

        assert len(states.batch_shape) == 1
        assert not torch.any(states.is_initial_state)

        states.forward_masks, states.backward_masks = correct_cast(
            states.forward_masks, states.backward_masks
        )

        incoming_log_flows = torch.full_like(
            states.backward_masks, -float("inf"), dtype=torch.float
        )
        outgoing_log_flows = torch.full_like(
            states.forward_masks, -float("inf"), dtype=torch.float
        )

        for action_idx in range(self.env.n_actions - 1):
            # TODO: can this be done in a vectorized way? Maybe by "repeating" the states and creating a big actions tensor?
            valid_backward_mask = states.backward_masks[:, action_idx]
            valid_forward_mask = states.forward_masks[:, action_idx]
            valid_backward_states = states[valid_backward_mask]
            valid_forward_states = states[valid_forward_mask]
            _, valid_backward_states.backward_masks = correct_cast(
                valid_backward_states.forward_masks,
                valid_backward_states.backward_masks,
            )
            backward_actions = torch.full_like(
                valid_backward_states.backward_masks[:, 0], action_idx, dtype=torch.long
            )

            valid_backward_states_parents = self.env.backward_step(
                valid_backward_states, backward_actions
            )

            incoming_log_flows[
                valid_backward_mask, action_idx
            ] = self.parametrization.logF(valid_backward_states_parents)[:, action_idx]
            outgoing_log_flows[
                valid_forward_mask, action_idx
            ] = self.parametrization.logF(valid_forward_states)[:, action_idx]

        # Now the exit action
        valid_forward_mask = states.forward_masks[:, -1]
        outgoing_log_flows[valid_forward_mask, -1] = self.parametrization.logF(
            states[valid_forward_mask]
        )[:, -1]

        log_incoming_flows = torch.logsumexp(incoming_log_flows, dim=-1)
        log_outgoing_flows = torch.logsumexp(outgoing_log_flows, dim=-1)

        return (log_incoming_flows - log_outgoing_flows).pow(2).mean()

    def reward_matching_loss(self, terminating_states: States) -> LossTensor:
        assert terminating_states.log_rewards is not None
        log_edge_flows = self.parametrization.logF(terminating_states)
        terminating_log_edge_flows = log_edge_flows[:, -1]
        log_rewards = terminating_states.log_rewards
        return (terminating_log_edge_flows - log_rewards).pow(2).mean()

    def __call__(self, states_tuple: Tuple[States, States]) -> LossTensor:
        intermediary_states, terminating_states = states_tuple
        fm_loss = self.flow_matching_loss(intermediary_states)
        rm_loss = self.reward_matching_loss(terminating_states)
        return fm_loss + self.alpha * rm_loss



"""
Copied and Adapted from https://github.com/Tikquuss/GflowNets_Tutorial
"""

from typing import ClassVar, Literal, Tuple, cast

import torch
from einops import rearrange
from gymnasium.spaces import Discrete
from torchtyping import TensorType

from gfn.containers.states import States
from gfn.envs.env import Env
from gfn.envs.preprocessors import (
    IdentityPreprocessor,
    KHotPreprocessor,
    OneHotPreprocessor,
)

# Typing
TensorLong = TensorType["batch_shape", torch.long]
TensorFloat = TensorType["batch_shape", torch.float]
TensorBool = TensorType["batch_shape", torch.bool]
ForwardMasksTensor = TensorType["batch_shape", "n_actions", torch.bool]
BackwardMasksTensor = TensorType["batch_shape", "n_actions - 1", torch.bool]
OneStateTensor = TensorType["state_shape", torch.float]
StatesTensor = TensorType["batch_shape", "state_shape", torch.float]

preprocessors_dict = {
    "KHot": KHotPreprocessor,
    "OneHot": OneHotPreprocessor,
    "Identity": IdentityPreprocessor,
}


class HyperGrid(Env):
    def __init__(
        self,
        ndim: int = 2,
        height: int = 4,
        R0: float = 0.1,
        R1: float = 0.5,
        R2: float = 2.0,
        reward_cos: bool = False,
        device_str: Literal["cpu", "cuda"] = "cpu",
        preprocessor_name: Literal["KHot", "OneHot", "Identity"] = "KHot",
    ):
        """HyperGrid environment from the GFlowNets paper.
        The states are represented as 1-d tensors of length `ndim` with values in
        {0, 1, ..., height - 1}.
        A preprocessor transforms the states to the input of the neural network,
        which can be a one-hot, a K-hot, or an identity encoding.

        Args:
            ndim (int, optional): dimension of the grid. Defaults to 2.
            height (int, optional): height of the grid. Defaults to 4.
            R0 (float, optional): reward parameter R0. Defaults to 0.1.
            R1 (float, optional): reward parameter R1. Defaults to 0.5.
            R2 (float, optional): reward parameter R1. Defaults to 2.0.
            reward_cos (bool, optional): Which version of the reward to use. Defaults to False.
            device_str (str, optional): "cpu" or "cuda". Defaults to "cpu".
            preprocessor_name (str, optional): "KHot" or "OneHot" or "Identity". Defaults to "KHot".
        """
        self.ndim = ndim
        self.height = height
        self.R0 = R0
        self.R1 = R1
        self.R2 = R2
        self.reward_cos = reward_cos

        s0 = torch.zeros(ndim, dtype=torch.long, device=torch.device(device_str))
        sf = torch.full(
            (ndim,), fill_value=-1, dtype=torch.long, device=torch.device(device_str)
        )

        action_space = Discrete(ndim + 1)

        if preprocessor_name == "Identity":
            preprocessor = IdentityPreprocessor(output_shape=(ndim,))
        elif preprocessor_name == "KHot":
            preprocessor = KHotPreprocessor(
                height=height, ndim=ndim, get_states_indices=self.get_states_indices
            )
        elif preprocessor_name == "OneHot":
            preprocessor = OneHotPreprocessor(
                n_states=self.n_states,
                get_states_indices=self.get_states_indices,
            )
        else:
            raise ValueError(f"Unknown preprocessor {preprocessor_name}")

        super().__init__(
            action_space=action_space,
            s0=s0,
            sf=sf,
            device_str=device_str,
            preprocessor=preprocessor,
        )

    def make_States_class(self) -> type[States]:
        "Creates a States class for this environment"
        env = self

        class HyperGridStates(States):

            state_shape: ClassVar[tuple[int, ...]] = (env.ndim,)
            s0 = env.s0
            sf = env.sf

            @classmethod
            def make_random_states_tensor(
                cls, batch_shape: Tuple[int, ...]
            ) -> StatesTensor:
                "Creates a batch of random states."
                states_tensor = torch.randint(
                    0, env.height, batch_shape + env.s0.shape, device=env.device
                )
                return states_tensor

            def make_masks(self) -> Tuple[ForwardMasksTensor, BackwardMasksTensor]:
                "Mask illegal (forward and backward) actions."
                forward_masks = torch.ones(
                    (*self.batch_shape, env.n_actions),
                    dtype=torch.bool,
                    device=env.device,
                )
                backward_masks = torch.ones(
                    (*self.batch_shape, env.n_actions - 1),
                    dtype=torch.bool,
                    device=env.device,
                )

                return forward_masks, backward_masks

            def update_masks(self) -> None:
                "Update the masks based on the current states."
                # The following two lines are for typing only.
                self.forward_masks = cast(ForwardMasksTensor, self.forward_masks)
                self.backward_masks = cast(BackwardMasksTensor, self.backward_masks)

                self.forward_masks[..., :-1] = self.states_tensor != env.height - 1
                self.backward_masks = self.states_tensor != 0

        return HyperGridStates

    def is_exit_actions(self, actions: TensorLong) -> TensorBool:
        return actions == self.action_space.n - 1

    def maskless_step(self, states: StatesTensor, actions: TensorLong) -> None:
        states.scatter_(-1, actions.unsqueeze(-1), 1, reduce="add")

    def maskless_backward_step(self, states: StatesTensor, actions: TensorLong) -> None:
        states.scatter_(-1, actions.unsqueeze(-1), -1, reduce="add")

    def true_reward(self, final_states: States) -> TensorFloat:
        final_states_raw = final_states.states_tensor
        R0, R1, R2 = (self.R0, self.R1, self.R2)
        ax = abs(final_states_raw / (self.height - 1) - 0.5)
        if not self.reward_cos:
            reward = (
                R0 + (0.25 < ax).prod(-1) * R1 + ((0.3 < ax) * (ax < 0.4)).prod(-1) * R2
            )
        else:
            pdf_input = ax * 5
            pdf = 1.0 / (2 * torch.pi) ** 0.5 * torch.exp(-(pdf_input**2) / 2)
            reward = R0 + ((torch.cos(ax * 50) + 1) * pdf).prod(-1) * R1
        return reward

    def log_reward(self, final_states: States) -> TensorFloat:
        return torch.log(self.true_reward(final_states))

    def get_states_indices(self, states: States) -> TensorLong:
        states_raw = states.states_tensor

        canonical_base = self.height ** torch.arange(
            self.ndim - 1, -1, -1, device=states_raw.device
        )
        indices = (canonical_base * states_raw).sum(-1).long()
        return indices

    def get_terminating_states_indices(self, states: States) -> TensorLong:
        return self.get_states_indices(states)

    @property
    def n_states(self) -> int:
        return self.height**self.ndim

    @property
    def n_terminating_states(self) -> int:
        return self.n_states

    @property
    def true_dist_pmf(self) -> torch.Tensor:
        all_states = self.all_states
        assert torch.all(
            self.get_states_indices(all_states)
            == torch.arange(self.n_states, device=self.device)
        )
        true_dist = self.reward(all_states)
        true_dist /= true_dist.sum()
        return true_dist

    @property
    def log_partition(self) -> float:
        grid = self.build_grid()
        rewards = self.reward(grid)
        return rewards.sum().log().item()

    def build_grid(self) -> States:
        "Utility function to build the complete grid"
        H = self.height
        ndim = self.ndim
        grid_shape = (H,) * ndim + (ndim,)  # (H, ..., H, ndim)
        grid = torch.zeros(grid_shape, device=self.device)
        for i in range(ndim):
            grid_i = torch.linspace(start=0, end=H - 1, steps=H)
            for _ in range(i):
                grid_i = grid_i.unsqueeze(1)
            grid[..., i] = grid_i

        rearrange_string = " ".join([f"n{i}" for i in range(1, ndim + 1)])
        rearrange_string += " ndim -> "
        rearrange_string += " ".join([f"n{i}" for i in range(ndim, 0, -1)])
        rearrange_string += " ndim"
        grid = rearrange(grid, rearrange_string).long()
        return self.States(grid)

    @property
    def all_states(self) -> States:
        grid = self.build_grid()
        flat_grid = rearrange(grid.states_tensor, "... ndim -> (...) ndim")
        return self.States(flat_grid)

    @property
    def terminating_states(self) -> States:
        return self.all_states



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional, Tuple, Union

import torch
from gymnasium.spaces import Discrete, Space
from torchtyping import TensorType

from gfn.containers.states import States, correct_cast
from gfn.envs.preprocessors import IdentityPreprocessor, Preprocessor

# Typing
TensorLong = TensorType["batch_shape", torch.long]
TensorFloat = TensorType["batch_shape", torch.float]
TensorBool = TensorType["batch_shape", torch.bool]
ForwardMasksTensor = TensorType["batch_shape", "n_actions", torch.bool]
BackwardMasksTensor = TensorType["batch_shape", "n_actions - 1", torch.bool]
OneStateTensor = TensorType["state_shape", torch.float]
StatesTensor = TensorType["batch_shape", "state_shape", torch.float]
PmfTensor = TensorType["n_states", torch.float]

NonValidActionsError = type("NonValidActionsError", (ValueError,), {})


class Env(ABC):
    """
    Base class for environments, showing which methods should be implemented.
    A common assumption for all environments is that all actions are discrete,
    represented by a number in {0, ..., n_actions - 1}, the last one being the
    exit action.
    """

    def __init__(
        self,
        action_space: Space,
        s0: OneStateTensor,
        sf: Optional[OneStateTensor] = None,
        device_str: Optional[str] = None,
        preprocessor: Optional[Preprocessor] = None,
    ):
        self.s0 = s0
        if sf is None:
            sf = torch.full(s0.shape, -float("inf"))
        self.sf = sf
        self.action_space = action_space
        self.device = torch.device(device_str) if device_str is not None else s0.device
        self.States = self.make_States_class()

        if preprocessor is None:
            preprocessor = IdentityPreprocessor(output_shape=tuple(s0.shape))

        self.preprocessor = preprocessor

    @abstractmethod
    def make_States_class(self) -> type[States]:
        "Returns a class that inherits from States and implements the environment-specific methods."
        pass

    @abstractmethod
    def is_exit_actions(self, actions: TensorLong) -> TensorBool:
        "Returns True if the action is an exit action."
        pass

    @abstractmethod
    def maskless_step(self, states: StatesTensor, actions: TensorLong) -> None:
        """Same as the step function, but without worrying whether or not the actions are valid, or masking."""
        pass

    @abstractmethod
    def maskless_backward_step(self, states: StatesTensor, actions: TensorLong) -> None:
        """Same as the backward_step function, but without worrying whether or not the actions are valid, or masking."""
        pass

    def reward(self, final_states: States) -> TensorFloat:
        """Either this or log_reward needs to be implemented."""
        return torch.exp(self.log_reward(final_states))

    def log_reward(self, final_states: States) -> TensorFloat:
        """Either this or reward needs to be implemented."""
        raise NotImplementedError("log_reward function not implemented")

    def get_states_indices(self, states: States) -> TensorLong:
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    def get_terminating_states_indices(self, states: States) -> TensorLong:
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    @property
    def n_actions(self) -> int:
        if isinstance(self.action_space, Discrete):
            return self.action_space.n
        else:
            raise NotImplementedError("Only discrete action spaces are supported")

    @property
    def n_states(self) -> int:
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    @property
    def n_terminating_states(self) -> int:
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    @property
    def true_dist_pmf(self) -> PmfTensor:
        "Returns a one-dimensional tensor representing the true distribution."
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    @property
    def log_partition(self) -> float:
        "Returns the logarithm of the partition function."
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    @property
    def all_states(self) -> States:
        """Returns a batch of all states for environments with enumerable states.
        The batch_shape should be (n_states,).
        This should satisfy:
        self.get_states_indices(self.all_states) == torch.arange(self.n_states)
        """
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    @property
    def terminating_states(self) -> States:
        """Returns a batch of all terminating states for environments with enumerable states.
        The batch_shape should be (n_terminating_states,).
        This should satisfy:
        self.get_terminating_states_indices(self.terminating_states) == torch.arange(self.n_terminating_states)
        """
        return NotImplementedError(
            "The environment does not support enumeration of states"
        )

    def reset(
        self, batch_shape: Union[int, Tuple[int]], random: bool = False
    ) -> States:
        "Instantiates a batch of initial states."
        if isinstance(batch_shape, int):
            batch_shape = (batch_shape,)
        return self.States.from_batch_shape(batch_shape=batch_shape, random=random)

    def step(
        self,
        states: States,
        actions: TensorLong,
    ) -> States:
        """Function that takes a batch of states and actions and returns a batch of next
        states and a boolean tensor indicating sink states in the new batch."""
        new_states = deepcopy(states)
        valid_states: TensorBool = ~states.is_sink_state
        valid_actions = actions[valid_states]

        if new_states.forward_masks is not None:
            new_forward_masks, _ = correct_cast(
                new_states.forward_masks, new_states.backward_masks
            )
            valid_states_masks = new_forward_masks[valid_states]
            valid_actions_bool = all(
                torch.gather(valid_states_masks, 1, valid_actions.unsqueeze(1))
            )
            if not valid_actions_bool:
                raise NonValidActionsError("Actions are not valid")

        new_sink_states = self.is_exit_actions(actions)
        new_states.states_tensor[new_sink_states] = self.sf
        new_sink_states = ~valid_states | new_sink_states

        not_done_states = new_states.states_tensor[~new_sink_states]
        not_done_actions = actions[~new_sink_states]

        self.maskless_step(not_done_states, not_done_actions)

        new_states.states_tensor[~new_sink_states] = not_done_states

        new_states.update_masks()
        return new_states

    def backward_step(self, states: States, actions: TensorLong) -> States:
        """Function that takes a batch of states and actions and returns a batch of next
        states and a boolean tensor indicating initial states in the new batch."""
        new_states = deepcopy(states)
        valid_states: TensorBool = ~new_states.is_initial_state
        valid_actions = actions[valid_states]

        if new_states.backward_masks is not None:
            _, new_backward_masks = correct_cast(
                new_states.forward_masks, new_states.backward_masks
            )
            valid_states_masks = new_backward_masks[valid_states]
            valid_actions_bool = all(
                torch.gather(valid_states_masks, 1, valid_actions.unsqueeze(1))
            )
            if not valid_actions_bool:
                raise NonValidActionsError("Actions are not valid")

        not_done_states = new_states.states_tensor[valid_states]
        self.maskless_backward_step(not_done_states, valid_actions)

        new_states.states_tensor[valid_states] = not_done_states

        new_states.update_masks()
        return new_states



from dataclasses import dataclass
from typing import List, Literal, Tuple

import torch
from torchtyping import TensorType

from gfn.containers import Trajectories
from gfn.estimators import LogStateFlowEstimator
from gfn.losses.base import PFBasedParametrization, TrajectoryDecomposableLoss
from gfn.samplers.actions_samplers import (
    BackwardDiscreteActionsSampler,
    DiscreteActionsSampler,
)

# Typing
ScoresTensor = TensorType[-1, float]
LossTensor = TensorType[0, float]
LogPTrajectoriesTensor = TensorType["max_length", "n_trajectories", float]


@dataclass
class SubTBParametrization(PFBasedParametrization):
    r"""
    Exactly the same as DBParametrization
    """
    logF: LogStateFlowEstimator


class SubTrajectoryBalance(TrajectoryDecomposableLoss):
    def __init__(
        self,
        parametrization: SubTBParametrization,
        log_reward_clip_min: float = -12,
        weighing: Literal[
            "DB",
            "ModifiedDB",
            "TB",
            "geometric",
            "equal",
            "geometric_within",
            "equal_within",
        ] = "geometric",
        lamda: float = 0.9,
        on_policy: bool = False,
    ):
        """
        :param parametrization: parametrization of the model
        :param log_reward_clip_min: minimum value of the log-reward. Log-Rewards lower than this value will be clipped to this value. Defaults to -12 (roughly log(1e-5)).
        :param weighing: how to weigh the different sub-trajectories of each trajectory.
            - "DB": Considers all one-step transitions of each trajectory in the batch and weighs them equally (regardless of the length of trajectory).
                    Should be equivalent to DetailedBalance loss.
            - "ModifiedDB": Considers all one-step transitions of each trajectory in the batch and weighs them inversely proportional to the trajectory length.
                    This ensures that the loss is not dominated by long trajectories. Each trajectory contributes equally to the loss.
            - "TB": Considers only the full trajectory. Should be equivalent to TrajectoryBalance loss.
            - "equal_within": Each sub-trajectory of each trajectory is weighed equally within the trajectory. Then each trajectory is weighed equally within the batch.
            - "equal": Each sub-trajectory of each trajectory is weighed equally within the set of all sub-trajectories.
            - "geometric_within": Each sub-trajectory of each trajectory is weighed proportionally to (lamda ** len(sub_trajectory)), within each trajectory.
            - "geometric": Each sub-trajectory of each trajectory is weighed proportionally to (lamda ** len(sub_trajectory)), within the set of all sub-trajectories.
        :param lamda: parameter for geometric weighing
        :param on_policy: whether the loss is computed on-policy (in which case the log probs stored in the trajectories are used) or off-policy
        """
        # Lamda is a discount factor for longer trajectories. The part of the loss
        # corresponding to sub-trajectories of length i is multiplied by lamda^i
        # where an edge is of length 1. As lamda approaches 1, each loss becomes equally weighted.
        self.parametrization = parametrization
        self.log_reward_clip_min = log_reward_clip_min
        self.actions_sampler = DiscreteActionsSampler(parametrization.logit_PF)
        self.backward_actions_sampler = BackwardDiscreteActionsSampler(
            parametrization.logit_PB
        )
        self.weighing = weighing
        self.lamda = lamda
        self.on_policy = on_policy

    def cumulative_logprobs(
        self,
        trajectories: Trajectories,
        log_p_trajectories: LogPTrajectoriesTensor,
    ) -> LogPTrajectoriesTensor:
        """
        :param trajectories: trajectories
        :param log_p_trajectories: log probabilities of each transition in each trajectory
        :return: cumulative sum of log probabilities of each trajectory
        """
        return torch.cat(
            (
                torch.zeros(
                    1, trajectories.n_trajectories, device=log_p_trajectories.device
                ),
                log_p_trajectories.cumsum(dim=0),
            ),
            dim=0,
        )

    def get_scores(
        self, trajectories: Trajectories
    ) -> Tuple[List[ScoresTensor], List[ScoresTensor]]:
        """
        Returns two elements:
        - A list of tensors, each of which representing the scores of all sub-trajectories of length k, for k in [1, ..., trajectories.max_length].
            where the score of a sub-trajectory tau is log P_F(tau) + log F(tau_0) - log P_B(tau) - log F(tau_{-1}). The shape of the k-th tensor
            is (trajectories.max_length - k + 1, trajectories.n_trajectories), k starting from 1.
        - A list of tensors representing what should be masked out in the each element of the first list, given that not all sub-trajectories
            of length k exist for each trajectory. The entries of those tensors are True if the corresponding sub-trajectory does not exist.
        """
        log_pf_trajectories, log_pb_trajectories = self.get_pfs_and_pbs(
            trajectories, fill_value=-float("inf"), no_pf=self.on_policy
        )
        if self.on_policy:
            log_pf_trajectories = trajectories.log_probs

        log_pf_trajectories_cum = self.cumulative_logprobs(
            trajectories, log_pf_trajectories
        )
        log_pb_trajectories_cum = self.cumulative_logprobs(
            trajectories, log_pb_trajectories
        )

        states = trajectories.states
        log_state_flows = torch.full_like(log_pf_trajectories, fill_value=-float("inf"))
        mask = ~states.is_sink_state
        valid_states = states[mask]
        log_state_flows[mask[:-1]] = self.parametrization.logF(valid_states).squeeze(-1)

        sink_states_mask = log_state_flows == -float("inf")
        is_terminal_mask = trajectories.actions == trajectories.env.n_actions - 1
        full_mask = sink_states_mask | is_terminal_mask

        flattening_masks = []
        scores = []
        for i in range(1, 1 + trajectories.max_length):
            current_log_state_flows = (
                log_state_flows if i == 1 else log_state_flows[: -(i - 1)]
            )

            preds = (
                log_pf_trajectories_cum[i:]
                - log_pf_trajectories_cum[:-i]
                + current_log_state_flows
            )

            targets = torch.full_like(preds, fill_value=-float("inf"))
            targets.T[is_terminal_mask[i - 1 :].T] = trajectories.log_rewards[trajectories.when_is_done >= i].clamp_min(self.log_reward_clip_min)  # type: ignore

            # For now, the targets contain the log-rewards of the ending sub trajectories
            # We need to add to that the log-probabilities of the backward actions up-to
            # the sub-trajectory's terminating state
            if i > 1:
                targets[is_terminal_mask[i - 1 :]] += (
                    log_pb_trajectories_cum[i - 1 :] - log_pb_trajectories_cum[: -i + 1]
                )[:-1][is_terminal_mask[i - 1 :]]

            # The following creates the targets for the non-finishing sub-trajectories
            targets[~full_mask[i - 1 :]] = (
                log_pb_trajectories_cum[i:] - log_pb_trajectories_cum[:-i]
            )[:-1][~full_mask[i - 1 : -1]] + log_state_flows[i:][~sink_states_mask[i:]]

            flattening_mask = trajectories.when_is_done.lt(
                torch.arange(
                    i,
                    trajectories.max_length + 1,
                    device=trajectories.when_is_done.device,
                ).unsqueeze(-1)
            )
            flat_preds = preds[~flattening_mask]
            flat_targets = targets[~flattening_mask]

            if torch.any(torch.isnan(flat_preds)):
                raise ValueError("NaN in preds")

            if torch.any(torch.isnan(flat_targets)):
                raise ValueError("NaN in targets")

            flattening_masks.append(flattening_mask)
            scores.append(preds - targets)

        return (
            scores,
            flattening_masks,
        )

    def __call__(self, trajectories: Trajectories) -> LossTensor:
        (
            scores,
            flattening_masks,
        ) = self.get_scores(trajectories)

        flattening_mask = torch.cat(flattening_masks)
        all_scores = torch.cat(scores, 0)

        # all_scores is a tensor of shape (max_length * (max_length + 1) / 2, n_trajectories)
        n_rows = int(trajectories.max_length * (1 + trajectories.max_length) / 2)

        if self.weighing == "equal_within":
            # the following tensor represents the inverse of how many sub-trajectories there are in each trajectory
            contributions = 2.0 / (
                trajectories.when_is_done * (trajectories.when_is_done + 1)
            )
            contributions = contributions / len(trajectories)
            # if we repeat the previous tensor, we get a tensor of shape (max_length * (max_length + 1) / 2, n_trajectories)
            # that we can multiply with all_scores to get a loss where each sub-trajectory is weighted equally within each trajectory
            contributions = contributions.repeat(n_rows, 1)
        elif self.weighing == "equal":
            n_sub_trajectories = int(
                (trajectories.when_is_done * (trajectories.when_is_done + 1) / 2)
                .sum()
                .item()
            )
            contributions = torch.ones(n_rows, len(trajectories)) / n_sub_trajectories
        elif self.weighing == "TB":
            # Each trajectory contributes one element to the loss, equally weighted
            contributions = torch.zeros_like(all_scores)
            indices = (
                trajectories.max_length * (trajectories.when_is_done - 1)
                - (trajectories.when_is_done - 1) * (trajectories.when_is_done - 2) / 2
            ).long()
            contributions.scatter_(0, indices.unsqueeze(0), 1)
            contributions = contributions / len(trajectories)
        elif self.weighing == "DB":
            # Longer trajectories contribute more to the loss
            return scores[0][~flattening_masks[0]].pow(2).mean()
        elif self.weighing == "ModifiedDB":
            # The following tensor represents the inverse of how many transitions there are in each trajectory
            contributions = 1.0 / trajectories.when_is_done
            contributions = contributions / len(trajectories)
            contributions = contributions.repeat(trajectories.max_length, 1)
            contributions = torch.cat(
                (
                    contributions,
                    torch.zeros(
                        (n_rows - trajectories.max_length, len(trajectories)),
                        device=contributions.device,
                    ),
                ),
                0,
            )

        elif self.weighing == "geometric_within":
            # the following tensor represents the weights given to each possible sub-trajectory length
            contributions = (
                self.lamda ** torch.arange(trajectories.max_length).double()
            ).float()
            contributions = contributions.unsqueeze(-1).repeat(1, len(trajectories))

            contributions = contributions.repeat_interleave(
                torch.arange(trajectories.max_length, 0, -1),
                dim=0,
                output_size=int(
                    trajectories.max_length * (trajectories.max_length + 1) / 2
                ),
            )
            r"""
            Now we need to divide each column by n + (n-1) lambda +...+ 1*lambda^{n-1}
            where n is the length of the trajectory corresponding to that column
            We can do it the ugly way, or using the cool identity:
            https://www.wolframalpha.com/input?i=sum%28%28n-i%29+*+lambda+%5Ei%2C+i%3D0..n%29
            """
            per_trajectory_denominator = (
                1.0
                / (1 - self.lamda) ** 2
                * (
                    self.lamda * (self.lamda ** trajectories.when_is_done.double() - 1)
                    + (1 - self.lamda) * trajectories.when_is_done.double()
                )
            ).float()
            contributions = contributions / per_trajectory_denominator
            contributions = contributions / len(trajectories)

        elif self.weighing == "geometric":
            # The position i of the following 1D tensor represents the number of sub-trajectories of length i in the batch
            # n_sub_trajectories = torch.maximum(
            #     trajectories.when_is_done - torch.arange(3).unsqueeze(-1),
            #     torch.tensor(0),
            # ).sum(1)

            # The following tensor's k-th entry represents the mean of all losses of sub-trajectories of length k
            per_length_losses = torch.stack(
                [
                    scores[~flattening_mask].pow(2).mean()
                    for scores, flattening_mask in zip(scores, flattening_masks)
                ]
            )
            ld = self.lamda
            weights = (
                (1 - ld)
                / (1 - ld**trajectories.max_length)
                * (
                    ld
                    ** torch.arange(
                        trajectories.max_length, device=per_length_losses.device
                    )
                )
            )
            assert (weights.sum() - 1.0).abs() < 1e-5, f"{weights.sum()}"
            return (per_length_losses * weights).sum()
        else:
            raise ValueError(f"Unknown weighing method {self.weighing}")

        flat_contributions = contributions[~flattening_mask]
        assert (
            flat_contributions.sum() - 1.0
        ).abs() < 1e-5, f"{flat_contributions.sum()}"
        losses = flat_contributions * all_scores[~flattening_mask].pow(2)
        return losses.sum()



from abc import ABC, abstractmethod
from typing import ClassVar, Literal, Tuple, cast

import torch
import torch.nn as nn
from gymnasium.spaces import Discrete
from torchtyping import TensorType

from gfn.containers.states import States
from gfn.envs.env import Env

# Typing
StatesTensor = TensorType["batch_shape", "state_shape", torch.float]
BatchTensor = TensorType["batch_shape"]
IsingJTensor = TensorType["state_shape", "state_shape", torch.float]
ForwardMasksTensor = TensorType["batch_shape", "n_actions", torch.bool]
BackwardMasksTensor = TensorType["batch_shape", "n_actions - 1", torch.bool]


class EnergyFunction(nn.Module, ABC):
    """Base class for energy functions"""

    @abstractmethod
    def forward(self, states: StatesTensor) -> BatchTensor:
        pass


class IsingModel(EnergyFunction):
    """Ising model energy function"""

    def __init__(self, J: IsingJTensor):
        super().__init__()
        self.J = J
        self.linear = nn.Linear(J.shape[0], 1, bias=False)
        self.linear.weight.data = J

    def forward(self, states: StatesTensor) -> BatchTensor:
        states = states.float()
        tmp = self.linear(states)
        return -(states * tmp).sum(-1)


class DiscreteEBMEnv(Env):
    """Environment for discrete energy-based models, based on https://arxiv.org/pdf/2202.01361.pdf"""

    def __init__(
        self,
        ndim: int,
        energy: EnergyFunction | None = None,
        alpha: float = 1.0,
        device_str: Literal["cpu", "cuda"] = "cpu",
    ):
        """Discrete EBM environment.

        Args:
            ndim (int, optional): dimension D of the sampling space {0, 1}^D.
            energy (EnergyFunction): energy function of the EBM. Defaults to None. If None, the Ising model with Identity matrix is used.
            alpha (float, optional): interaction strength the EBM. Defaults to 1.0.
            device_str (str, optional): "cpu" or "cuda". Defaults to "cpu".
        """
        self.ndim = ndim

        s0 = torch.full((ndim,), -1, dtype=torch.long, device=torch.device(device_str))
        sf = torch.full((ndim,), 2, dtype=torch.long, device=torch.device(device_str))

        if energy is None:
            energy = IsingModel(
                torch.ones((ndim, ndim), device=torch.device(device_str))
            )
        self.energy: EnergyFunction = energy
        self.alpha = alpha

        action_space = Discrete(
            2 * ndim + 1
        )  # the last action is the exit action that is only available for complete states
        # Action i in [0, ndim - 1] corresponds to replacing s[i] with 0
        # Action i in [ndim, 2 * ndim - 1] corresponds to replacing s[i - ndim] with 1

        super().__init__(action_space=action_space, s0=s0, sf=sf)

    def make_States_class(self) -> type[States]:
        env = self

        class DiscreteEBMStates(States):
            state_shape: ClassVar[tuple[int, ...]] = (env.ndim,)
            s0 = env.s0
            sf = env.sf

            @classmethod
            def make_random_states_tensor(
                cls, batch_shape: Tuple[int, ...]
            ) -> StatesTensor:
                return torch.randint(
                    -1,
                    2,
                    batch_shape + (env.ndim,),
                    dtype=torch.long,
                    device=env.device,
                )

            def make_masks(
                self,
            ) -> Tuple[ForwardMasksTensor, BackwardMasksTensor]:
                forward_masks = torch.zeros(
                    self.batch_shape + (env.n_actions,),
                    device=env.device,
                    dtype=torch.bool,
                )
                backward_masks = torch.zeros(
                    self.batch_shape + (env.n_actions - 1,),
                    device=env.device,
                    dtype=torch.bool,
                )

                return forward_masks, backward_masks

            def update_masks(self) -> None:
                # The following two lines are for typing only.
                self.forward_masks = cast(ForwardMasksTensor, self.forward_masks)
                self.backward_masks = cast(BackwardMasksTensor, self.backward_masks)

                self.forward_masks[..., : env.ndim] = self.states_tensor == -1
                self.forward_masks[..., env.ndim : 2 * env.ndim] = (
                    self.states_tensor == -1
                )
                self.forward_masks[..., -1] = torch.all(
                    self.states_tensor != -1, dim=-1
                )
                self.backward_masks[..., : env.ndim] = self.states_tensor == 0
                self.backward_masks[..., env.ndim : 2 * env.ndim] = (
                    self.states_tensor == 1
                )

        return DiscreteEBMStates

    def is_exit_actions(self, actions: BatchTensor) -> BatchTensor:
        return actions == self.n_actions - 1

    def maskless_step(self, states: StatesTensor, actions: BatchTensor) -> None:
        # First, we select that actions that replace a -1 with a 0
        mask_0 = actions < self.ndim
        states[mask_0] = states[mask_0].scatter(-1, actions[mask_0].unsqueeze(-1), 0)
        # Then, we select that actions that replace a -1 with a 1
        mask_1 = (actions >= self.ndim) & (actions < 2 * self.ndim)
        states[mask_1] = states[mask_1].scatter(
            -1, (actions[mask_1] - self.ndim).unsqueeze(-1), 1
        )

    def maskless_backward_step(
        self, states: StatesTensor, actions: BatchTensor
    ) -> None:
        states.scatter_(-1, actions.unsqueeze(-1).fmod(self.ndim), -1)

    def log_reward(self, final_states: States) -> BatchTensor:
        raw_states = final_states.states_tensor
        canonical = 2 * raw_states - 1
        return -self.alpha * self.energy(canonical)

    def get_states_indices(self, states: States) -> BatchTensor:
        """The chosen encoding is the following: -1 -> 0, 0 -> 1, 1 -> 2, then we convert to base 3"""
        states_raw = states.states_tensor
        canonical_base = 3 ** torch.arange(self.ndim - 1, -1, -1, device=self.device)
        return (states_raw + 1).mul(canonical_base).sum(-1).long()

    def get_terminating_states_indices(self, states: States) -> BatchTensor:
        states_raw = states.states_tensor
        canonical_base = 2 ** torch.arange(self.ndim - 1, -1, -1, device=self.device)
        return (states_raw).mul(canonical_base).sum(-1).long()

    @property
    def n_states(self) -> int:
        return 3**self.ndim

    @property
    def n_terminating_states(self) -> int:
        return 2**self.ndim

    @property
    def all_states(self) -> States:
        # This is brute force !
        digits = torch.arange(3, device=self.device)
        all_states = torch.cartesian_prod(*[digits] * self.ndim)
        all_states = all_states - 1
        return self.States(all_states)

    @property
    def terminating_states(self) -> States:
        digits = torch.arange(2, device=self.device)
        all_states = torch.cartesian_prod(*[digits] * self.ndim)
        return self.States(all_states)

    @property
    def true_dist_pmf(self) -> torch.Tensor:
        true_dist = self.reward(self.terminating_states)
        return true_dist / true_dist.sum()

    @property
    def log_partition(self) -> float:
        log_rewards = self.log_reward(self.terminating_states)
        return torch.logsumexp(log_rewards, -1).item()



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

import torch
from torchtyping import TensorType

if TYPE_CHECKING:
    from gfn.envs import Env
    from gfn.containers.states import States

from gfn.containers.base import Container

# Typing  -- n_transitions is either int or Tuple[int]
LongTensor = TensorType["n_transitions", torch.long]
BoolTensor = TensorType["n_transitions", torch.bool]
FloatTensor = TensorType["n_transitions", torch.float]
PairFloatTensor = TensorType["n_transitions", 2, torch.float]


class Transitions(Container):
    def __init__(
        self,
        env: Env,
        states: States | None = None,
        actions: LongTensor | None = None,
        is_done: BoolTensor | None = None,
        next_states: States | None = None,
        is_backward: bool = False,
        log_rewards: FloatTensor | None = None,
        log_probs: FloatTensor | None = None,
    ):
        """Container for transitions.

        Args:
            env (Env): Environment
            states (States, optional): States object with uni-dimensional batch_shape, representing the parents of the transitions. Defaults to None.
            actions (LongTensor, optional): Actions chosen at the parents of each transitions. Defaults to None.
            is_done (BoolTensor, optional): Whether the action is the exit action. Defaults to None.
            next_states (States, optional): States object with uni-dimensional batch_shape, representing the children of the transitions. Defaults to None.
            is_backward (bool, optional): Whether the transitions are backward transitions (i.e. next_states is the parent of states). Defaults to False.
            log_rewards (FloatTensor1D, optional): The log-rewards of the transitions (using a default value like -1 for non-terminating transitions). Defaults to None.
            log_probs (FloatTensor1D, optional): The log-probabilities of the actions. Defaults to None.

        When states and next_states are not None, the Transitions is an empty container that can be populated on the go.
        """
        self.env = env
        self.is_backward = is_backward
        self.states = (
            states
            if states is not None
            else env.States.from_batch_shape(batch_shape=(0,))
        )
        assert len(self.states.batch_shape) == 1
        self.actions = (
            actions
            if actions is not None
            else torch.full(size=(0,), fill_value=-1, dtype=torch.long)
        )
        self.is_done = (
            is_done
            if is_done is not None
            else torch.full(size=(0,), fill_value=False, dtype=torch.bool)
        )
        self.next_states = (
            next_states
            if next_states is not None
            else env.States.from_batch_shape(batch_shape=(0,))
        )
        assert (
            len(self.next_states.batch_shape) == 1
            and self.states.batch_shape == self.next_states.batch_shape
        )

        self._log_rewards = log_rewards

        self.log_probs = log_probs if log_probs is not None else torch.zeros(0)

    @property
    def n_transitions(self) -> int:
        return self.states.batch_shape[0]

    def __len__(self) -> int:
        return self.n_transitions

    def __repr__(self):
        states_tensor = self.states.states_tensor
        next_states_tensor = self.next_states.states_tensor

        states_repr = ",\t".join(
            [
                f"{str(state.numpy())} -> {str(next_state.numpy())}"
                for state, next_state in zip(states_tensor, next_states_tensor)
            ]
        )
        return (
            f"Transitions(n_transitions={self.n_transitions}, "
            f"transitions={states_repr}, actions={self.actions}, "
            f"is_done={self.is_done})"
        )

    @property
    def last_states(self) -> States:
        "Get the last states, i.e. terminating states"
        return self.states[self.is_done]

    @property
    def log_rewards(self) -> FloatTensor | None:
        if self._log_rewards is not None:
            return self._log_rewards
        if self.is_backward:
            return None
        else:
            log_rewards = torch.full(
                (self.n_transitions,),
                fill_value=-1.0,
                dtype=torch.float,
                device=self.states.device,
            )
            try:
                log_rewards[self.is_done] = self.env.log_reward(self.last_states)
            except NotImplementedError:
                log_rewards[self.is_done] = torch.log(self.env.reward(self.last_states))
            return log_rewards

    @property
    def all_log_rewards(self) -> PairFloatTensor:
        """This is applicable to environments where all states are terminating.
        This function evaluates the rewards for all transitions that do not end in the sink state.
        This is useful for the Modified Detailed Balance loss."""
        if self.is_backward:
            raise NotImplementedError("Not implemented for backward transitions")
        is_sink_state = self.next_states.is_sink_state
        log_rewards = torch.full(
            (self.n_transitions, 2),
            fill_value=-1.0,
            dtype=torch.float,
            device=self.states.device,
        )
        try:
            log_rewards[~is_sink_state, 0] = self.env.log_reward(
                self.states[~is_sink_state]
            )
            log_rewards[~is_sink_state, 1] = self.env.log_reward(
                self.next_states[~is_sink_state]
            )
        except NotImplementedError:
            log_rewards[~is_sink_state, 0] = torch.log(
                self.env.reward(self.states[~is_sink_state])
            )
            log_rewards[~is_sink_state, 1] = torch.log(
                self.env.reward(self.next_states[~is_sink_state])
            )
        return log_rewards

    def __getitem__(self, index: int | Sequence[int]) -> Transitions:
        "Access particular transitions of the batch."
        if isinstance(index, int):
            index = [index]
        states = self.states[index]
        actions = self.actions[index]
        is_done = self.is_done[index]
        next_states = self.next_states[index]
        log_rewards = (
            self._log_rewards[index] if self._log_rewards is not None else None
        )
        log_probs = self.log_probs[index]
        return Transitions(
            env=self.env,
            states=states,
            actions=actions,
            is_done=is_done,
            next_states=next_states,
            is_backward=self.is_backward,
            log_rewards=log_rewards,
            log_probs=log_probs,
        )

    def extend(self, other: Transitions) -> None:
        "Extend the Transitions object with another Transitions object."
        self.states.extend(other.states)
        self.actions = torch.cat((self.actions, other.actions), dim=0)
        self.is_done = torch.cat((self.is_done, other.is_done), dim=0)
        self.next_states.extend(other.next_states)
        if self._log_rewards is not None and other._log_rewards is not None:
            self._log_rewards = torch.cat(
                (self._log_rewards, other._log_rewards), dim=0
            )
        else:
            self._log_rewards = None
        self.log_probs = torch.cat((self.log_probs, other.log_probs), dim=0)



from abc import ABC, abstractmethod
from typing import Tuple

import torch
from torch.distributions import Categorical
from torchtyping import TensorType

from gfn.containers.states import States, correct_cast
from gfn.estimators import LogEdgeFlowEstimator, LogitPBEstimator, LogitPFEstimator

# Typing
Tensor2D = TensorType["batch_size", "n_actions"]
Tensor2D2 = TensorType["batch_size", "n_steps"]
Tensor1D = TensorType["batch_size", torch.long]


class ActionsSampler(ABC):
    """
    Base class for action sampling methods.
    """

    @abstractmethod
    def sample(self, states: States) -> Tuple[Tensor1D, Tensor1D]:
        """
        Args:
            states (States): A batch of states.

        Returns:
            Tuple[Tensor[batch_size], Tensor[batch_size]]: A tuple of tensors containing the log probabilities of the sampled actions, and the sampled actions.
        """
        pass


class BackwardActionsSampler(ActionsSampler):
    """
    Base class for backward action sampling methods.
    """

    pass


class DiscreteActionsSampler(ActionsSampler):
    """
    For Discrete environments.
    """

    def __init__(
        self,
        estimator: LogitPFEstimator | LogEdgeFlowEstimator,
        temperature: float = 1.0,
        sf_bias: float = 0.0,
        epsilon: float = 0.0,
    ) -> None:
        """Implements a method that samples actions from any given batch of states.

        Args:
            temperature (float, optional): scalar to divide the logits by before softmax. Defaults to 1.0.
            sf_bias (float, optional): scalar to subtract from the exit action logit before dividing by temperature. Defaults to 0.0.
            epsilon (float, optional): with probability epsilon, a random action is chosen. Defaults to 0.0.
        """
        self.estimator = estimator
        self.temperature = temperature
        self.sf_bias = sf_bias
        self.epsilon = epsilon

    def get_raw_logits(self, states: States) -> Tensor2D:
        """
        This is before illegal actions are masked out and the exit action is biased.
        Should be used for Discrete action spaces only.

        Returns:
            Tensor2D: A 2D tensor of shape (batch_size, n_actions) containing the logits for each action in each state in the batch.
        """
        logits = self.estimator(states)
        return logits

    def get_logits(self, states: States) -> Tensor2D:
        """Transforms the raw logits by masking illegal actions.

        Raises:
            ValueError: if one of the resulting logits is NaN.

        Returns:
            Tensor2D: A 2D tensor of shape (batch_size, n_actions) containing the transformed logits.
        """
        logits = self.get_raw_logits(states)

        if torch.any(torch.all(torch.isnan(logits), 1)):
            raise ValueError("NaNs in estimator")
        states.forward_masks, _ = correct_cast(
            states.forward_masks, states.backward_masks
        )
        logits[~states.forward_masks] = -float("inf")
        return logits

    def get_probs(
        self,
        states: States,
    ) -> Tensor2D:
        """
        Returns:
            The probabilities of each action in each state in the batch.
        """
        logits = self.get_logits(states)
        logits[..., -1] -= self.sf_bias
        probs = torch.softmax(logits / self.temperature, dim=-1)
        return probs

    def sample(self, states: States) -> Tuple[Tensor1D, Tensor1D]:
        probs = self.get_probs(states)
        states.forward_masks, _ = correct_cast(
            states.forward_masks, states.backward_masks
        )
        if self.epsilon > 0:
            uniform_dist = (
                states.forward_masks.float()
                / states.forward_masks.sum(dim=-1, keepdim=True).float()
            )
            probs = (1 - self.epsilon) * probs + self.epsilon * uniform_dist
        dist = Categorical(probs=probs)
        with torch.no_grad():
            actions = dist.sample()
        actions_log_probs = dist.log_prob(actions)

        return actions_log_probs, actions


class BackwardDiscreteActionsSampler(DiscreteActionsSampler, BackwardActionsSampler):
    """
    For sampling backward actions in discrete environments.
    """

    def __init__(
        self,
        estimator: LogitPBEstimator,
        temperature: float = 1.0,
        epsilon: float = 0.0,
    ) -> None:
        """s_f is not biased in the backward sampler."""
        super().__init__(
            estimator, temperature=temperature, sf_bias=0.0, epsilon=epsilon
        )

    def get_logits(self, states: States) -> Tensor2D:
        logits = self.get_raw_logits(states)
        if torch.any(torch.all(torch.isnan(logits), 1)):
            raise ValueError("NaNs in estimator")
        _, states.backward_masks = correct_cast(
            states.forward_masks, states.backward_masks
        )
        logits[~states.backward_masks] = -float("inf")
        return logits

    def get_probs(self, states: States) -> Tensor2D:
        logits = self.get_logits(states)
        probs = torch.softmax(logits / self.temperature, dim=-1)
        # The following line is hack that works: when probs are nan, it means
        # that the state is already done (usually during backward sampling).
        # In which case, any action can be passed to the backward_step function
        # making the state stay at s_0
        probs = probs.nan_to_num(nan=1.0 / probs.shape[-1])
        return probs



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Sequence

if TYPE_CHECKING:
    from gfn.envs import Env

import torch
from torchtyping import TensorType

# Typing  --- n_transitions is an int
Tensor2D = TensorType["max_length", "n_trajectories", torch.long]
Tensor2D2 = TensorType["n_trajectories", "shape"]
Tensor1D = TensorType["n_trajectories", torch.long]
FloatTensor1D = TensorType["n_trajectories", torch.float]


class Container(ABC):
    "Base class for states containers (states, transitions, or trajectories)"

    @abstractmethod
    def __len__(self) -> int:
        "Returns the number of elements in the container"
        pass

    @abstractmethod
    def __getitem__(self, index: int | Sequence[int]) -> Container:
        "Subsets the container"
        pass

    @abstractmethod
    def extend(self, other: Container) -> None:
        "Extends the current container"
        pass

    def sample(self, n_samples: int) -> Container:
        "Samples a subset of the container"
        return self[torch.randperm(len(self))[:n_samples]]

    def save(self, path: str) -> None:
        "Saves the container to a file"
        for key, val in self.__dict__.items():
            if isinstance(val, Env):
                continue
            elif isinstance(val, Container):
                val.save(os.path.join(path, key))
            elif isinstance(val, torch.Tensor):
                torch.save(val, os.path.join(path, key + ".pt"))
            else:
                raise ValueError(f"Unexpected {key} of type {type(val)}")

    def load(self, path: str) -> None:
        "Loads the container from a file, overwriting the current container"
        for key, val in self.__dict__.items():
            if isinstance(val, Env):
                continue
            elif isinstance(val, Container):
                val.load(os.path.join(path, key))
            elif isinstance(val, torch.Tensor):
                self.__dict__[key] = torch.load(os.path.join(path, key + ".pt"))
            else:
                raise ValueError(f"Unexpected {key} of type {type(val)}")



"""
Implementations of the [Trajectory Balance loss](https://arxiv.org/abs/2201.13259)
and the [Log Partition Variance loss](https://arxiv.org/abs/2302.05446).
"""

from dataclasses import dataclass

import torch
from torchtyping import TensorType

from gfn.containers import Trajectories
from gfn.estimators import LogZEstimator
from gfn.losses.base import PFBasedParametrization, TrajectoryDecomposableLoss
from gfn.samplers.actions_samplers import (
    BackwardDiscreteActionsSampler,
    DiscreteActionsSampler,
)

# Typing
ScoresTensor = TensorType["n_trajectories", float]
LossTensor = TensorType[0, float]


@dataclass
class TBParametrization(PFBasedParametrization):
    r"""
    $\mathcal{O}_{PFZ} = \mathcal{O}_1 \times \mathcal{O}_2 \times \mathcal{O}_3$, where
    $\mathcal{O}_1 = \mathbb{R}$ represents the possible values for logZ,
    and $\mathcal{O}_2$ is the set of forward probability functions consistent with the DAG.
    $\mathcal{O}_3$ is the set of backward probability functions consistent with the DAG, or a singleton
    thereof, if self.logit_PB is a fixed LogitPBEstimator.
    Useful for the Trajectory Balance Loss.
    """
    logZ: LogZEstimator


class TrajectoryBalance(TrajectoryDecomposableLoss):
    def __init__(
        self,
        parametrization: TBParametrization,
        log_reward_clip_min: float = -12,
        on_policy: bool = False,
    ):
        """Loss object to evaluate the TB loss on a batch of trajectories.

        Args:
            log_reward_clip_min (float, optional): minimal value to clamp the reward to. Defaults to -12 (roughly log(1e-5)).
            on_policy (bool, optional): If True, the log probs stored in the trajectories are used. Which should be faster than
                                        reevaluating them. Defaults to False.
        """
        self.parametrization = parametrization
        self.log_reward_clip_min = log_reward_clip_min
        self.actions_sampler = DiscreteActionsSampler(parametrization.logit_PF)
        self.backward_actions_sampler = BackwardDiscreteActionsSampler(
            parametrization.logit_PB
        )
        self.on_policy = on_policy

    def __call__(self, trajectories: Trajectories) -> LossTensor:
        _, _, scores = self.get_trajectories_scores(trajectories)
        loss = (scores + self.parametrization.logZ.tensor).pow(2).mean()
        if torch.isnan(loss):
            raise ValueError("loss is nan")

        return loss


class LogPartitionVarianceLoss(TrajectoryDecomposableLoss):
    def __init__(
        self,
        parametrization: PFBasedParametrization,
        log_reward_clip_min: float = -12,
        on_policy: bool = False,
    ):
        """Loss object to evaluate the Log Partition Variance Loss (Section 3.2 of
        [ROBUST SCHEDULING WITH GFLOWNETS](https://arxiv.org/abs/2302.05446))

        Args:
            on_policy (bool, optional): If True, the log probs stored in the trajectories are used. Which should be faster than
                                        reevaluating them. Defaults to False.
        """
        self.parametrization = parametrization
        self.log_reward_clip_min = log_reward_clip_min
        self.actions_sampler = DiscreteActionsSampler(parametrization.logit_PF)
        self.backward_actions_sampler = BackwardDiscreteActionsSampler(
            parametrization.logit_PB
        )

        self.on_policy = on_policy

    def __call__(self, trajectories: Trajectories) -> LossTensor:
        _, _, scores = self.get_trajectories_scores(trajectories)
        loss = (scores - scores.mean()).pow(2).mean()
        if torch.isnan(loss):
            raise ValueError("loss is nan")

        return loss



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Tuple

import torch
from torchtyping import TensorType

from gfn.containers.states import States, correct_cast
from gfn.containers.trajectories import Trajectories
from gfn.containers.transitions import Transitions
from gfn.distributions import (
    EmpiricalTrajectoryDistribution,
    TrajectoryBasedTerminatingStateDistribution,
    TrajectoryDistribution,
)
from gfn.envs import Env
from gfn.estimators import LogitPBEstimator, LogitPFEstimator
from gfn.samplers import DiscreteActionsSampler, TrajectoriesSampler

# Typing
LogPTrajectoriesTensor = TensorType["max_length", "n_trajectories", float]
ScoresTensor = TensorType["n_trajectories", float]


@dataclass
class Parametrization(ABC):
    """
    Abstract Base Class for Flow Parametrizations,
    as defined in Sec. 3 of GFlowNets Foundations.
    All attributes should be estimators, and should either have a GFNModule or attribute called `module`,
    or torch.Tensor attribute called `tensor` with requires_grad=True.
    """

    @abstractmethod
    def Pi(self, env: Env, n_samples: int, **kwargs) -> TrajectoryDistribution:
        pass

    def P_T(
        self, env: Env, n_samples: int, **kwargs
    ) -> TrajectoryBasedTerminatingStateDistribution:
        return TrajectoryBasedTerminatingStateDistribution(
            self.Pi(env, n_samples, **kwargs)
        )

    @property
    def parameters(self) -> dict:
        """
        Return a dictionary of all parameters of the parametrization.
        Note that there might be duplicate parameters (e.g. when two NNs share parameters),
        in which case the optimizer should take as input set(self.parameters.values()).
        """
        # TODO: use parameters of the fields instead, loop through them here
        parameters_dict = {}
        for name, estimator in self.__dict__.items():
            parameters_dict.update(
                {
                    f"{name}_{key}": value
                    for key, value in estimator.named_parameters().items()
                }
            )
        return parameters_dict

    def save_state_dict(self, path: str):
        for name, estimator in self.__dict__.items():
            torch.save(estimator.named_parameters(), os.path.join(path, name + ".pt"))

    def load_state_dict(self, path: str):
        for name, estimator in self.__dict__.items():
            estimator.load_state_dict(torch.load(os.path.join(path, name + ".pt")))


@dataclass
class PFBasedParametrization(Parametrization, ABC):
    r"Base class for parametrizations that explicitly uses $P_F$"
    logit_PF: LogitPFEstimator
    logit_PB: LogitPBEstimator

    def Pi(
        self, env: Env, n_samples: int = 1000, **actions_sampler_kwargs
    ) -> TrajectoryDistribution:
        actions_sampler = DiscreteActionsSampler(
            self.logit_PF, **actions_sampler_kwargs
        )
        trajectories_sampler = TrajectoriesSampler(env, actions_sampler)
        trajectories = trajectories_sampler.sample_trajectories(
            n_trajectories=n_samples
        )
        return EmpiricalTrajectoryDistribution(trajectories)


class Loss(ABC):
    "Abstract Base Class for all GFN Losses"

    def __init__(self, parametrization: Parametrization):
        self.parametrization = parametrization

    @abstractmethod
    def __call__(self, *args, **kwargs) -> TensorType[0, float]:
        pass


class EdgeDecomposableLoss(Loss, ABC):
    @abstractmethod
    def __call__(self, edges: Transitions) -> TensorType[0, float]:
        pass


class StateDecomposableLoss(Loss, ABC):
    @abstractmethod
    def __call__(self, states_tuple: Tuple[States, States]) -> TensorType[0, float]:
        """Unlike the GFlowNets Foundations paper, we allow more flexibility by passing a tuple of states,
        the first one being the internal states of the trajectories (i.e. non-terminal states), and the second one
        being the terminal states of the trajectories. If these two are not handled differently, then they should be
        concatenated together."""
        pass


class TrajectoryDecomposableLoss(Loss, ABC):
    def get_pfs_and_pbs(
        self,
        trajectories: Trajectories,
        fill_value: float = 0.0,
        temperature: float = 1.0,
        epsilon=0.0,
        no_pf: bool = False,
    ) -> Tuple[LogPTrajectoriesTensor | None, LogPTrajectoriesTensor]:
        """Evaluate log_pf and log_pb for each action in each trajectory in the batch.
        This is useful when the policy used to sample the trajectories is different from the one used to evaluate the loss.

        Args:
            trajectories (Trajectories): Trajectories to evaluate.
            fill_value (float, optional): Value to use for invalid states (i.e. s_f that is added to shorter trajectories). Defaults to 0.0.

            The next parameters correspond to how the actions_sampler evaluates each action.
            temperature (float, optional): Temperature to use for the softmax. Defaults to 1.0.
            epsilon (float, optional): Epsilon to use for the softmax. Defaults to 0.0.
            no_pf (bool, optional): Whether to evaluate log_pf as well. Defaults to False.

        Raises:
            ValueError: if the trajectories are backward.

        Returns:
            Tuple[LogPTrajectoriesTensor | None, LogPTrajectoriesTensor]: A tuple of float tensors of shape (max_length, n_trajectories) containing the log_pf and log_pb for each action in each trajectory. The first one can be None.
        """
        # fill value is the value used for invalid states (sink state usually)
        if trajectories.is_backward:
            raise ValueError("Backward trajectories are not supported")

        valid_states = trajectories.states[~trajectories.states.is_sink_state]
        valid_actions = trajectories.actions[trajectories.actions != -1]

        # uncomment next line for debugging
        # assert trajectories.states.is_sink_state[:-1].equal(trajectories.actions == -1)

        if valid_states.batch_shape != tuple(valid_actions.shape):
            raise AssertionError("Something wrong happening with log_pf evaluations")

        valid_states.forward_masks, valid_states.backward_masks = correct_cast(
            valid_states.forward_masks, valid_states.backward_masks
        )
        log_pf_trajectories = None
        if not no_pf:
            valid_pf_logits = self.actions_sampler.get_logits(valid_states)
            valid_pf_logits = valid_pf_logits / temperature
            valid_log_pf_all = valid_pf_logits.log_softmax(dim=-1)
            valid_log_pf_all = (
                1 - epsilon
            ) * valid_log_pf_all + epsilon * valid_states.forward_masks.float() / valid_states.forward_masks.sum(
                dim=-1, keepdim=True
            )
            valid_log_pf_actions = torch.gather(
                valid_log_pf_all, dim=-1, index=valid_actions.unsqueeze(-1)
            ).squeeze(-1)
            log_pf_trajectories = torch.full_like(
                trajectories.actions, fill_value=fill_value, dtype=torch.float
            )
            log_pf_trajectories[trajectories.actions != -1] = valid_log_pf_actions

        valid_pb_logits = self.backward_actions_sampler.get_logits(
            valid_states[~valid_states.is_initial_state]
        )
        valid_log_pb_all = valid_pb_logits.log_softmax(dim=-1)
        non_exit_valid_actions = valid_actions[
            valid_actions != trajectories.env.n_actions - 1
        ]
        valid_log_pb_actions = torch.gather(
            valid_log_pb_all, dim=-1, index=non_exit_valid_actions.unsqueeze(-1)
        ).squeeze(-1)
        log_pb_trajectories = torch.full_like(
            trajectories.actions, fill_value=fill_value, dtype=torch.float
        )
        log_pb_trajectories_slice = torch.full_like(
            valid_actions, fill_value=fill_value, dtype=torch.float
        )
        log_pb_trajectories_slice[
            valid_actions != trajectories.env.n_actions - 1
        ] = valid_log_pb_actions
        log_pb_trajectories[trajectories.actions != -1] = log_pb_trajectories_slice

        return log_pf_trajectories, log_pb_trajectories

    def get_trajectories_scores(
        self, trajectories: Trajectories
    ) -> Tuple[ScoresTensor, ScoresTensor, ScoresTensor]:

        log_pf_trajectories, log_pb_trajectories = self.get_pfs_and_pbs(
            trajectories, no_pf=self.on_policy
        )
        if self.on_policy:
            log_pf_trajectories = trajectories.log_probs

        assert log_pf_trajectories is not None
        log_pf_trajectories = log_pf_trajectories.sum(dim=0)
        log_pb_trajectories = log_pb_trajectories.sum(dim=0)

        log_rewards = trajectories.log_rewards.clamp_min(self.log_reward_clip_min)  # type: ignore

        return (
            log_pf_trajectories,
            log_pb_trajectories,
            log_pf_trajectories - log_pb_trajectories - log_rewards,
        )

    @abstractmethod
    def __call__(self, trajectories: Trajectories) -> TensorType[0, float]:
        pass



from dataclasses import dataclass

import torch
from torchtyping import TensorType

from gfn.containers import Transitions
from gfn.estimators import LogStateFlowEstimator
from gfn.losses.base import EdgeDecomposableLoss, PFBasedParametrization
from gfn.samplers.actions_samplers import (
    BackwardDiscreteActionsSampler,
    DiscreteActionsSampler,
)

# Typing
ScoresTensor = TensorType["n_transitions", float]
LossTensor = TensorType[0, float]


@dataclass
class DBParametrization(PFBasedParametrization):
    r"""
    Corresponds to $\mathcal{O}_{PF} = \mathcal{O}_1 \times \mathcal{O}_2 \times \mathcal{O}_3$, where
    $\mathcal{O}_1$ is the set of functions from the internal states (no $s_f$)
    to $\mathbb{R}^+$ (which we parametrize with logs, to avoid the non-negativity constraint),
    and $\mathcal{O}_2$ is the set of forward probability functions consistent with the DAG.
    $\mathcal{O}_3$ is the set of backward probability functions consistent with the DAG, or a singleton
    thereof, if self.logit_PB is a fixed LogitPBEstimator.
    Useful for the Detailed Balance Loss.
    """
    logF: LogStateFlowEstimator


class DetailedBalance(EdgeDecomposableLoss):
    def __init__(self, parametrization: DBParametrization, on_policy: bool = False):
        "If on_policy is True, the log probs stored in the transitions are used."
        self.parametrization = parametrization
        self.actions_sampler = DiscreteActionsSampler(parametrization.logit_PF)
        self.backward_actions_sampler = BackwardDiscreteActionsSampler(
            parametrization.logit_PB
        )
        self.on_policy = on_policy

    def get_scores(self, transitions: Transitions):
        if transitions.is_backward:
            raise ValueError("Backward transitions are not supported")
        states = transitions.states
        actions = transitions.actions

        # uncomment next line for debugging
        # assert transitions.states.is_sink_state.equal(transitions.actions == -1)

        if states.batch_shape != tuple(actions.shape):
            raise ValueError("Something wrong happening with log_pf evaluations")
        if self.on_policy:
            valid_log_pf_actions = transitions.log_probs
        else:
            valid_pf_logits = self.actions_sampler.get_logits(states)
            valid_log_pf_all = valid_pf_logits.log_softmax(dim=-1)
            valid_log_pf_actions = torch.gather(
                valid_log_pf_all, dim=-1, index=actions.unsqueeze(-1)
            ).squeeze(-1)

        valid_log_F_s = self.parametrization.logF(states).squeeze(-1)

        preds = valid_log_pf_actions + valid_log_F_s

        targets = torch.zeros_like(preds)

        # uncomment next line for debugging
        # assert transitions.next_states.is_sink_state.equal(transitions.is_done)

        # automatically removes invalid transitions (i.e. s_f -> s_f)
        valid_next_states = transitions.next_states[~transitions.is_done]
        non_exit_actions = actions[actions != transitions.env.n_actions - 1]
        valid_pb_logits = self.backward_actions_sampler.get_logits(valid_next_states)
        valid_log_pb_all = valid_pb_logits.log_softmax(dim=-1)
        valid_log_pb_actions = torch.gather(
            valid_log_pb_all, dim=-1, index=non_exit_actions.unsqueeze(-1)
        ).squeeze(-1)

        valid_transitions_is_done = transitions.is_done[
            ~transitions.states.is_sink_state
        ]

        valid_log_F_s_next = self.parametrization.logF(valid_next_states).squeeze(-1)
        targets[~valid_transitions_is_done] = valid_log_pb_actions
        log_pb_actions = targets.clone()
        targets[~valid_transitions_is_done] += valid_log_F_s_next
        assert transitions.log_rewards is not None
        valid_transitions_log_rewards = transitions.log_rewards[
            ~transitions.states.is_sink_state
        ]
        targets[valid_transitions_is_done] = valid_transitions_log_rewards[
            valid_transitions_is_done
        ]

        scores = preds - targets

        return (valid_log_pf_actions, log_pb_actions, scores)

    def __call__(self, transitions: Transitions) -> LossTensor:
        _, _, scores = self.get_scores(transitions)
        loss = torch.mean(scores**2)

        if torch.isnan(loss):
            raise ValueError("loss is nan")

        return loss

    def get_modified_scores(self, transitions: Transitions) -> ScoresTensor:
        "DAG-GFN-style detailed balance, for when all states are connected to the sink"
        if transitions.is_backward:
            raise ValueError("Backward transitions are not supported")
        mask = ~transitions.next_states.is_sink_state
        states = transitions.states[mask]
        valid_next_states = transitions.next_states[mask]
        actions = transitions.actions[mask]
        all_log_rewards = transitions.all_log_rewards[mask]

        valid_pf_logits = self.actions_sampler.get_logits(states)
        valid_log_pf_all = valid_pf_logits.log_softmax(dim=-1)
        valid_log_pf_actions = torch.gather(
            valid_log_pf_all, dim=-1, index=actions.unsqueeze(-1)
        ).squeeze(-1)
        valid_log_pf_s_exit = valid_log_pf_all[:, -1]

        # The following two lines are slightly inefficient, given that most
        # next_states are also states, for which we already did a forward pass.
        valid_log_pf_s_prime_all = self.actions_sampler.get_logits(
            valid_next_states
        ).log_softmax(dim=-1)
        valid_log_pf_s_prime_exit = valid_log_pf_s_prime_all[:, -1]

        valid_pb_logits = self.backward_actions_sampler.get_logits(valid_next_states)
        valid_log_pb_all = valid_pb_logits.log_softmax(dim=-1)
        valid_log_pb_actions = torch.gather(
            valid_log_pb_all, dim=-1, index=actions.unsqueeze(-1)
        ).squeeze(-1)

        preds = all_log_rewards[:, 0] + valid_log_pf_actions + valid_log_pf_s_prime_exit
        targets = all_log_rewards[:, 1] + valid_log_pb_actions + valid_log_pf_s_exit

        scores = preds - targets
        if torch.any(torch.isinf(scores)):
            raise ValueError("scores contains inf")

        return scores



from .base import (
    EdgeDecomposableLoss,
    Loss,
    Parametrization,
    PFBasedParametrization,
    StateDecomposableLoss,
    TrajectoryDecomposableLoss,
)
from .detailed_balance import DBParametrization, DetailedBalance
from .flow_matching import FlowMatching, FMParametrization
from .sub_trajectory_balance import SubTBParametrization, SubTrajectoryBalance
from .trajectory_balance import (
    LogPartitionVarianceLoss,
    TBParametrization,
    TrajectoryBalance,
)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gfn',
 'gfn.containers',
 'gfn.envs',
 'gfn.envs.preprocessors',
 'gfn.losses',
 'gfn.samplers']

package_data = \
{'': ['*']}

install_requires = \
['einops>=0.6.1',
 'gymnasium>=0.28.1',
 'numpy>=1.21.2',
 'torch>=1.9.0',
 'torchtyping>=0.1.4']

extras_require = \
{'dev': ['pre-commit',
         'pytest',
         'renku-sphinx-theme',
         'sphinx_rtd_theme',
         'sphinx-autoapi',
         'sphinx-math-dollar',
         'sphinx',
         'tox',
         'black==22.3.0',
         'myst-parser'],
 'scripts': ['tqdm', 'wandb', 'simple-parsing==0.0.20']}

setup_kwargs = {
    'name': 'torchgfn',
    'version': '0.1.3',
    'description': 'A torch inplementation of GFlowNets',
    'long_description': '<p align="center">\n    <a>\n\t    <img src=\'https://img.shields.io/badge/python-3.10%2B-blueviolet\' alt=\'Python\' />\n\t</a>\n\t<a href=\'https://gfn.readthedocs.io/en/latest/?badge=latest\'>\n    \t<img src=\'https://readthedocs.org/projects/gfn/badge/?version=latest\' alt=\'Documentation Status\' />\n\t</a>\n    <a>\n\t    <img src=\'https://img.shields.io/badge/code%20style-black-black\' />\n\t</a>\n</p>\n\n</p>\n<p align="center">\n  <a href="https://gfn.readthedocs.io/en/latest/">Documentation</a> ~ <a href="https://github.com/saleml/gfn">Code</a>\n</p>\n\n# gfn: a Python package for GFlowNets\n\n\n\n## Installing the packages\n\nThe codebase requires python >= 3.10\n\n```bash\ngit clone https://github.com/saleml/gfn.git\nconda create -n gfn python=3.10\nconda activate gfn\ncd gfn\npip install .\n```\n\nOptionally, to run scripts, and for [wandb](https://wandb.ai) logging\n\n```bash\npip install .[scripts]\nwandb login\n```\n\n## About this repo\n\nThis repo serves the purpose of fast prototyping [GFlowNet](https://arxiv.org/abs/2111.09266) related algorithms. It decouples the environment definition, the sampling process, and the parametrization used for the GFN loss.\n\nAn example script is provided [here](https://github.com/saleml/gfn/blob/master/scripts/train.py). To run the code, use one of the following:\n\n```bash\npython train.py --env HyperGrid --env.ndim 4 --env.height 8 --n_iterations 100000 --loss TB\npython train.py --env DiscreteEBM --env.ndim 4 --env.alpha 0.5 --n_iterations 10000 --batch_size 64 --temperature 2.\npython train.py --env HyperGrid --env.ndim 2 --env.height 64 --n_iterations 100000 --loss DB --replay_buffer_size 1000 --logit_PB.module_name Uniform --optim sgd --optim.lr 5e-3\npython train.py --env HyperGrid --env.ndim 4 --env.height 8 --env.R0 0.01 --loss FM --optim adam --optim.lr 1e-4\n```\n\n### Example, in a few lines\n\n(⬇️ This example require the [`tqdm`](https://github.com/tqdm/tqdm) package to run. `pip install tqdm` or install all extra requirements with `pip install .[scripts]`)\n\n```python\nimport torch\nfrom tqdm import tqdm\n\nfrom gfn import LogitPBEstimator, LogitPFEstimator, LogZEstimator\nfrom gfn.envs import HyperGrid\nfrom gfn.losses import TBParametrization, TrajectoryBalance\nfrom gfn.samplers import DiscreteActionsSampler, TrajectoriesSampler\n\nif __name__ == "__main__":\n\n    env = HyperGrid(ndim=4, height=8, R0=0.01)  # Grid of size 8x8x8x8\n\n    logit_PF = LogitPFEstimator(env=env, module_name="NeuralNet")\n    logit_PB = LogitPBEstimator(\n        env=env,\n        module_name="NeuralNet",\n        torso=logit_PF.module.torso,  # To share parameters between PF and PB\n    )\n    logZ = LogZEstimator(torch.tensor(0.0))\n\n    parametrization = TBParametrization(logit_PF, logit_PB, logZ)\n\n    actions_sampler = DiscreteActionsSampler(estimator=logit_PF)\n    trajectories_sampler = TrajectoriesSampler(env=env, actions_sampler=actions_sampler)\n\n    loss_fn = TrajectoryBalance(parametrization=parametrization)\n\n    params = [\n        {\n            "params": [\n                val for key, val in parametrization.parameters.items() if "logZ" not in key\n            ],\n            "lr": 0.001,\n        },\n        {"params": [val for key, val in parametrization.parameters.items() if "logZ" in key], "lr": 0.1},\n    ]\n    optimizer = torch.optim.Adam(params)\n\n    for i in (pbar := tqdm(range(1000))):\n        trajectories = trajectories_sampler.sample(n_trajectories=16)\n        optimizer.zero_grad()\n        loss = loss_fn(trajectories)\n        loss.backward()\n        optimizer.step()\n        if i % 25 == 0:\n            pbar.set_postfix({"loss": loss.item()})\n\n```\n\n## Contributing\n\nBefore the first commit:\n\n```bash\npip install .[dev]\npre-commit install\npre-commit run --all-files\n```\n\nRun `pre-commit` after staging, and before committing. Make sure all the tests pass (By running `pytest`).\nThe codebase uses `black` formatter.\n\nTo make the docs locally:\n```bash\ncd docs\nmake html\nopen build/html/index.html\n```\n\n## Details about the codebase\n\n### Defining an environment\n\nA pointed DAG environment (or GFN environment, or environment for short) is a representation for the pointed DAG. The abstract class [Env](https://github.com/saleml/gfn/blob/master/src/gfn/envs/env.py) specifies the requirements for a valid environment definition. To obtain such a representation, the environment needs to specify the following attributes, properties, or methods:\n\n- The `action_space`. Which should be a `gymnasium.spaces.Discrete` object for discrete environments. The last action should correspond to the exit action.\n- The initial state `s_0`, as a `torch.Tensor` of arbitrary dimension.\n- (Optional) The sink state `s_f`, as a `torch.Tensor` of the same shape as `s_0`, used to represent complete trajectories only (within a batch of trajectories of different lengths), and never processed by any model. If not specified, it is set to `torch.full_like(s_0, -float(\'inf\'))`.\n- The method `make_States_class` that creates a subclass of [States](https://github.com/saleml/gfn/blob/master/src/gfn/containers/states.py). The instances of the resulting class should represent a batch of states of arbitrary shape, which is useful to define a trajectory, or a batch of trajectories. `s_0` and `s_f`, along with a tuple called `state_shape` should be defined as class variables, and the subclass (of `States`) should implement masking methods, that specify which actions are possible, in a discrete environment.\n- The methods `maskless_step` and `maskless_backward_step` that specify how an action changes a state (going forward and backward). These functions do not need to handle masking, checking whether actions are allowed, checking whether a state is the sink state, etc... These checks are handled in `Env.step` and `Env.backward_step`\n- The `log_reward` function that assigns a nonnegative reward to every terminating state (i.e. state with all $s_f$ as a child in the DAG). If `log_reward` is not implemented, `reward` needs to be.\n\nIf the states (as represented in the `States` class) need to be transformed to another format before being processed (by neural networks for example), then the environment should define a `preprocessor` attribute, which should be an instance of the [base preprocessor class](https://github.com/saleml/gfn/blob/master/src/gfn/envs/preprocessors/base.py). If no preprocessor is defined, the states are used as is (actually transformed using  [`IdentityPreprocessor`](https://github.com/saleml/gfn/blob/master/src/gfn/envs/preprocessors/base.py), which transforms the state tensors to `FloatTensor`s). Implementing your own preprocessor requires defining the `preprocess` function, and the `output_shape` attribute, which is a tuple representing the shape of *one* preprocessed state.\n\nOptionally, you can define a static `get_states_indices` method that assigns a unique integer number to each state if the environment allows it, and a `n_states` property that returns an integer representing the number of states (excluding $s_f$) in the environment. `get_terminating_states_indices` can also be implemented and serves the purpose of uniquely identifying terminating states of the environment.\n\nFor more details, take a look at [HyperGrid](https://github.com/saleml/gfn/blob/master/src/gfn/envs/hypergrid.py), an environment where all states are terminating states, or at [DiscreteEBM](https://github.com/saleml/gfn/blob/master/src/gfn/envs/discrete_ebm.py), where all trajectories are of the same length but only some states are terminating.\n\n### Other containers\n\nBesides the `States` class, other containers of states are available:\n\n- [Transitions](https://github.com/saleml/gfn/blob/master/src/gfn/containers/transitions.py), representing a batch of transitions $s \\rightarrow s\'$.\n- [Trajectories](https://github.com/saleml/gfn/blob/master/src/gfn/containers/trajectories.py), representing a batch of complete trajectories $\\tau = s_0 \\rightarrow s_1 \\rightarrow \\dots \\rightarrow s_n \\rightarrow s_f$.\n\nThese containers can either be instantiated using a `States` object, or can be initialized as empty containers that can be populated on the fly, allowing the usage of [ReplayBuffer](https://github.com/saleml/gfn/blob/master/src/gfn/containers/replay_buffer.py)s.\n\nThey inherit from the base `Container` [class](https://github.com/saleml/gfn/blob/master/src/gfn/containers/base.py), indicating some helpful methods.\n\nIn most cases, one needs to sample complete trajectories. From a batch of trajectories, a batch of states and batch of transitions can be defined using `Trajectories.to_transitions()` and `Trajectories.to_states()`. These exclude meaningless transitions and states that were added to the batch of trajectories to allow for efficient batching.\n\n### Estimators and Modules\n\nTraining GFlowNets requires one or multiple estimators. As of now, only discrete environments are handled. All estimators are subclasses of [FunctionEstimator](https://github.com/saleml/gfn/blob/master/src/gfn/estimators.py), implementing a `__call__` function that takes as input a batch of [States](https://github.com/saleml/gfn/blob/master/src/gfn/containers/states.py).\n\n- [LogEdgeFlowEstimator](https://github.com/saleml/gfn/blob/master/src/gfn/estimators.py). It outputs a `(*batch_shape, n_actions)` tensor representing $\\log F(s \\rightarrow s\')$, including when $s\' = s_f$.\n- [LogStateFlowEstimator](https://github.com/saleml/gfn/blob/master/src/gfn/estimators.py). It outputs a `(*batch_shape, 1)` tensor representing $\\log F(s)$. When used with `forward_looking=True`, $\\log F(s)$ is parametrized as the sum of a function approximator and $\\log R(s)$ - which is only possible for environments where all states are terminating.\n- [LogitPFEstimator](https://github.com/saleml/gfn/blob/master/src/gfn/estimators.py). It outputs a `(*batch_shape, n_actions)` tensor representing $logit(s\' \\mid s)$, such that $P_F(s\' \\mid s) = softmax_{s\'}\\ logit(s\' \\mid s)$, including when $s\' = s_f$.\n- [LogitPBEstimator](https://github.com/saleml/gfn/blob/master/src/gfn/estimators.py). It outputs a `(*batch_shape, n_actions - 1)` tensor representing $logit(s\' \\mid s)$, such that $P_B(s\' \\mid s) = softmax_{s\'}\\ logit(s\' \\mid s)$.\n\nDefining an estimator requires the environment, and a [module](https://github.com/saleml/gfn/blob/master/src/gfn/modules.py) instance. Modules inherit from the [GFNModule](https://github.com/saleml/gfn/blob/master/src/gfn/modules.py) class, which can be seen as an extension of `torch.nn.Module`. Alternatively, a module is created by providing which module type to use (e.g. "NeuralNet" or "Uniform" or "Zero"). A Basic MLP is provided as the [NeuralNet](https://github.com/saleml/gfn/blob/master/src/gfn/modules.py) class, but any function approximator should be possible.\n\nSaid differently, a `States` object is first transformed via the environment\'s preprocessor to a `(*batch_shape, *output_shape)` float tensor. The preprocessor\'s output shape should match the module input shape (if any). The preprocessed states are then passed as inputs to the module, returning the desired output (either flows or probabilities over children in the DAG).\n\nEach module has a `named_parameters` functions that returns a dictionary of the learnable parameters. This attribute is transferred to the corresponding estimator.\n\nAdditionally, a [LogZEstimator](https://github.com/saleml/gfn/blob/master/src/gfn/estimators.py) is provided, which is a container for a scalar tensor representing $\\log Z$, the log-partition function, useful for the Trajectory Balance loss for example. This estimator also has a `named_parameters` function.\n\n### Samplers\n\nAn [ActionsSampler](https://github.com/saleml/gfn/blob/master/src/gfn/samplers/actions_samplers.py) object defines how actions are sampled at each state of the DAG. As of now, only [DiscreteActionsSampler](https://github.com/saleml/gfn/blob/master/src/gfn/samplers/actions_samplers.py)s are implemented. The require an estimator (of $P_F$, $P_B$, or edge flows) defining the action probabilities. These estimators can contain any type of modules (including random action sampling for example). A [BackwardDiscreteActionsSampler](https://github.com/saleml/gfn/blob/master/src/gfn/samplers/actions_samplers.py) class is provided to sample parents of a state, which is helpful to sample trajectories starting from their last states.\n\nThey are at the core of [TrajectoriesSampler](https://github.com/saleml/gfn/blob/master/src/gfn/samplers/trajectories_sampler.py)s, which implements the `sample_trajectories` method, that sample a batch of trajectories starting from a given set of initial states or starting from $s_0$.\n\n### Losses\n\nGFlowNets can be trained with different losses, each of which requires a different parametrization. A parametrization is a dataclass, which can be seen as a container of different estimators. Each parametrization defines a distribution over trajectories, via the `parametrization.Pi` method, and a distribution over terminating states, via the `parametrization.P_T` method. Both distributions should be instances of the classes defined [here](https://github.com/saleml/gfn/blob/master/src/gfn/distributions.py).\n\nThe base classes for losses and parametrizations are provided [here](https://github.com/saleml/gfn/blob/master/src/gfn/losses/base.py).\n\nCurrently, the implemented losses are:\n\n- Flow Matching\n- Detailed Balance\n- Trajectory Balance\n- Sub-Trajectory Balance. By default, each sub-trajectory is weighted geometrically (within the trajectory) depending on its length. This corresponds to the strategy defined [here](https://www.semanticscholar.org/reader/f2c32fe3f7f3e2e9d36d833e32ec55fc93f900f5). Other strategies exist and are implemented [here](https://github.com/saleml/gfn/blob/master/src/gfn/losses/sub_trajectory_balance.py).\n- Log Partition Variance loss. Introduced [here](https://arxiv.org/abs/2302.05446)\n\n### Solving for the flows using Dynamic Programming\n\nA simple script that propagates trajectories rewards through the DAG to define edge flows in a deterministic way (by visiting each edge once only) is provided [here](https://github.com/saleml/gfn/blob/master/scripts/dynamic_programming.py). Do not use the script on large environments !\n',
    'author': 'Salem Lahou',
    'author_email': 'salemlahlou9@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from typing import Literal, Optional

import torch
import torch.nn as nn
from torchtyping import TensorType

# Typing

InputTensor = TensorType["batch_shape", "input_shape", float]
OutputTensor = TensorType["batch_shape", "output_dim", float]


class GFNModule(ABC):
    """Abstract Base Class for all functions/approximators/estimators used.
    Each module takes a preprocessed tensor as input, and outputs a tensor of logits,
    or log flows. The input dimension of the module (e.g. Neural network), is deduced
    from the environment's preprocessor's output dimension"""

    @property
    @abstractmethod
    def output_dim(self) -> int:
        """Output dimension of the module"""
        pass

    def named_parameters(self) -> dict:
        """Returns a dictionary of all (learnable) parameters of the module. Not needed
        for NeuralNet modules, given that those inherit this function from nn.Module"""
        return {}

    @abstractmethod
    def __call__(self, preprocessed_states: InputTensor) -> OutputTensor:
        pass

    @abstractmethod
    def load_state_dict(self, state_dict: dict):
        pass


class NeuralNet(nn.Module, GFNModule):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dim: Optional[int] = 256,
        n_hidden_layers: Optional[int] = 2,
        activation_fn: Optional[Literal["relu", "tanh"]] = "relu",
        torso: Optional[nn.Module] = None,
    ):
        """Implements a basic MLP.

        Args:
            input_dim (int): input dimension
            output_dim (int): output dimension
            hidden_dim (Optional[int], optional): Number of units per hidden layer. Defaults to 256.
            n_hidden_layers (Optional[int], optional): Number of hidden layers. Defaults to 2.
            activation_fn (Optional[Literal[relu, tanh]], optional): Activation function. Defaults to "relu".
            torso (Optional[nn.Module], optional): If provided, this module will be used as the torso of the network (i.e. all layers except last layer). Defaults to None.
        """
        super().__init__()
        self._output_dim = output_dim

        if torso is None:
            assert (
                n_hidden_layers is not None and n_hidden_layers >= 0
            ), "n_hidden_layers must be >= 0"
            assert activation_fn is not None, "activation_fn must be provided"
            activation = nn.ReLU if activation_fn == "relu" else nn.Tanh
            self.torso = [nn.Linear(input_dim, hidden_dim), activation()]
            for _ in range(n_hidden_layers - 1):
                self.torso.append(nn.Linear(hidden_dim, hidden_dim))
                self.torso.append(activation())
            self.torso = nn.Sequential(*self.torso)
            self.torso.hidden_dim = hidden_dim
        else:
            self.torso = torso
        self.last_layer = nn.Linear(self.torso.hidden_dim, output_dim)
        self.device = None

    def forward(self, preprocessed_states: InputTensor) -> OutputTensor:
        if self.device is None:
            self.device = preprocessed_states.device
            self.to(self.device)
        logits = self.torso(preprocessed_states)
        logits = self.last_layer(logits)
        return logits

    @property
    def output_dim(self) -> int:
        return self._output_dim


class Tabular(GFNModule):
    def __init__(self, n_states: int, output_dim: int) -> None:
        """Implements a tabular function approximator. Only compatible with the EnumPreprocessor.

        Args:
            n_states (int): Number of states in the environment.
            output_dim (int): Output dimension.
        """
        self._output_dim = output_dim

        self.logits = torch.zeros(
            (n_states, output_dim),
            dtype=torch.float,
            requires_grad=True,
        )

        self.device = None

    @property
    def output_dim(self) -> int:
        return self._output_dim

    def __call__(self, preprocessed_states: InputTensor) -> OutputTensor:
        if self.device is None:
            self.device = preprocessed_states.device
            self.logits = self.logits.to(self.device)
        assert preprocessed_states.dtype == torch.long
        outputs = self.logits[preprocessed_states.squeeze(-1)]
        return outputs

    def named_parameters(self) -> dict:
        return {"logits": self.logits}

    def load_state_dict(self, state_dict: dict):
        self.logits = state_dict["logits"]


class ZeroGFNModule(GFNModule):
    def __init__(self, output_dim: int) -> None:
        """Implements a zero function approximator, i.e. a function that always outputs 0.

        Args:
            output_dim (int): Output dimension.
        """
        self._output_dim = output_dim

    def __call__(self, preprocessed_states: InputTensor) -> OutputTensor:
        out = torch.zeros(*preprocessed_states.shape[:-1], self.output_dim).to(
            preprocessed_states.device
        )
        return out

    @property
    def output_dim(self) -> int:
        return self._output_dim

    def load_state_dict(self, state_dict: dict):
        pass


class Uniform(ZeroGFNModule):
    """Use this module for uniform policies for example. This is because logits = 0 is equivalent to uniform policy"""

    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

if TYPE_CHECKING:
    from gfn.envs import Env
    from gfn.containers.states import States

import torch
from torchtyping import TensorType

from gfn.containers.base import Container
from gfn.containers.transitions import Transitions

# Typing  --- n_transitions is an int
Tensor2D = TensorType["max_length", "n_trajectories", torch.long]
FloatTensor2D = TensorType["max_length", "n_trajectories", torch.float]
Tensor2D2 = TensorType["n_trajectories", "shape"]
Tensor1D = TensorType["n_trajectories", torch.long]
FloatTensor1D = TensorType["n_trajectories", torch.float]


class Trajectories(Container):
    def __init__(
        self,
        env: Env,
        states: States | None = None,
        actions: Tensor2D | None = None,
        when_is_done: Tensor1D | None = None,
        is_backward: bool = False,
        log_rewards: FloatTensor1D | None = None,
        log_probs: FloatTensor2D | None = None,
    ) -> None:
        """Container for complete trajectories (starting in s_0 and ending in s_f).
        Trajectories are represented as a States object with bi-dimensional batch shape.
        The first dimension represents the time step, the second dimension represents the trajectory index.
        Because different trajectories may have different lengths, shorter trajectories are padded with
        the tensor representation of the terminal state (s_f or s_0 depending on the direction of the trajectory), and
        actions is appended with -1's.
        The actions are represented as a two dimensional tensor with the first dimension representing the time step
        and the second dimension representing the trajectory index.
        The when_is_done tensor represents the time step at which each trajectory ends.


        Args:
            env (Env): The environment in which the trajectories are defined.
            states (States, optional): The states of the trajectories. Defaults to None.
            actions (Tensor2D, optional): The actions of the trajectories. Defaults to None.
            when_is_done (Tensor1D, optional): The time step at which each trajectory ends. Defaults to None.
            is_backward (bool, optional): Whether the trajectories are backward or forward. Defaults to False.
            log_rewards (FloatTensor1D, optional): The log_rewards of the trajectories. Defaults to None.
            log_probs (FloatTensor2D, optional): The log probabilities of the trajectories' actions. Defaults to None.

        If states is None, then the states are initialized to an empty States object, that can be populated on the fly.
        If log_rewards is None, then `env.log_reward` is used to compute the rewards, at each call of self.log_rewards
        """
        self.env = env
        self.is_backward = is_backward
        self.states = (
            states
            if states is not None
            else env.States.from_batch_shape(batch_shape=(0, 0))
        )
        assert len(self.states.batch_shape) == 2
        self.actions = (
            actions
            if actions is not None
            else torch.full(size=(0, 0), fill_value=-1, dtype=torch.long)
        )
        self.when_is_done = (
            when_is_done
            if when_is_done is not None
            else torch.full(size=(0,), fill_value=-1, dtype=torch.long)
        )
        self._log_rewards = log_rewards
        self.log_probs = (
            log_probs
            if log_probs is not None
            else torch.full(size=(0, 0), fill_value=0, dtype=torch.float)
        )

    def __repr__(self) -> str:
        states = self.states.states_tensor.transpose(0, 1)
        assert states.ndim == 3
        trajectories_representation = ""
        for traj in states[:10]:
            one_traj_repr = []
            for step in traj:
                one_traj_repr.append(str(step.numpy()))
                if step.equal(self.env.s0 if self.is_backward else self.env.sf):
                    break
            trajectories_representation += "-> ".join(one_traj_repr) + "\n"
        return (
            f"Trajectories(n_trajectories={self.n_trajectories}, max_length={self.max_length}, First 10 trajectories:"
            + f"states=\n{trajectories_representation}, actions=\n{self.actions.transpose(0, 1)[:10].numpy()}, "
            + f"when_is_done={self.when_is_done[:10].numpy()})"
        )

    @property
    def n_trajectories(self) -> int:
        return self.states.batch_shape[1]

    def __len__(self) -> int:
        return self.n_trajectories

    @property
    def max_length(self) -> int:
        if len(self) == 0:
            return 0

        return self.actions.shape[0]

    @property
    def last_states(self) -> States:
        return self.states[self.when_is_done - 1, torch.arange(self.n_trajectories)]

    @property
    def log_rewards(self) -> FloatTensor1D | None:
        if self._log_rewards is not None:
            assert self._log_rewards.shape == (self.n_trajectories,)
            return self._log_rewards
        if self.is_backward:
            return None
        try:
            return self.env.log_reward(self.last_states)
        except NotImplementedError:
            return torch.log(self.env.reward(self.last_states))

    def __getitem__(self, index: int | Sequence[int]) -> Trajectories:
        "Returns a subset of the `n_trajectories` trajectories."
        if isinstance(index, int):
            index = [index]
        when_is_done = self.when_is_done[index]
        new_max_length = when_is_done.max().item() if len(when_is_done) > 0 else 0
        states = self.states[:, index]
        actions = self.actions[:, index]
        log_probs = self.log_probs[:, index]
        states = states[: 1 + new_max_length]
        actions = actions[:new_max_length]
        log_probs = log_probs[:new_max_length]
        log_rewards = (
            self._log_rewards[index] if self._log_rewards is not None else None
        )

        return Trajectories(
            env=self.env,
            states=states,
            actions=actions,
            when_is_done=when_is_done,
            is_backward=self.is_backward,
            log_rewards=log_rewards,
            log_probs=log_probs,
        )

    def extend(self, other: Trajectories) -> None:
        """Extend the trajectories with another set of trajectories."""
        self.extend_actions(required_first_dim=max(self.max_length, other.max_length))
        other.extend_actions(required_first_dim=max(self.max_length, other.max_length))

        self.states.extend(other.states)
        self.actions = torch.cat((self.actions, other.actions), dim=1)
        self.when_is_done = torch.cat((self.when_is_done, other.when_is_done), dim=0)
        self.log_probs = torch.cat((self.log_probs, other.log_probs), dim=1)

        if self._log_rewards is not None and other._log_rewards is not None:
            self._log_rewards = torch.cat(
                (self._log_rewards, other._log_rewards), dim=0
            )
        else:
            self._log_rewards = None

    def extend_actions(self, required_first_dim: int) -> None:
        """Extends the actions and log_probs along the first dimension by by adding -1s as necessary.
        This is useful for extending trajectories of different lengths."""
        if self.max_length >= required_first_dim:
            return
        self.actions = torch.cat(
            (
                self.actions,
                torch.full(
                    size=(
                        required_first_dim - self.actions.shape[0],
                        self.n_trajectories,
                    ),
                    fill_value=-1,
                    dtype=torch.long,
                ),
            ),
            dim=0,
        )
        self.log_probs = torch.cat(
            (
                self.log_probs,
                torch.full(
                    size=(
                        required_first_dim - self.log_probs.shape[0],
                        self.n_trajectories,
                    ),
                    fill_value=0,
                    dtype=torch.float,
                ),
            ),
            dim=0,
        )

    @staticmethod
    def revert_backward_trajectories(trajectories: Trajectories) -> Trajectories:

        assert trajectories.is_backward
        new_actions = torch.full_like(trajectories.actions, -1)
        new_actions = torch.cat(
            [new_actions, torch.full((1, len(trajectories)), -1)], dim=0
        )
        new_states = trajectories.env.sf.repeat(
            trajectories.when_is_done.max() + 1, len(trajectories), 1
        )
        new_when_is_done = trajectories.when_is_done + 1
        for i in range(len(trajectories)):
            new_actions[trajectories.when_is_done[i], i] = (
                trajectories.env.n_actions - 1
            )
            new_actions[: trajectories.when_is_done[i], i] = trajectories.actions[
                : trajectories.when_is_done[i], i
            ].flip(0)
            new_states[
                : trajectories.when_is_done[i] + 1, i
            ] = trajectories.states.states_tensor[
                : trajectories.when_is_done[i] + 1, i
            ].flip(
                0
            )
        new_states = trajectories.env.States(new_states)
        return Trajectories(
            env=trajectories.env,
            states=new_states,
            actions=new_actions,
            log_probs=trajectories.log_probs,
            when_is_done=new_when_is_done,
            is_backward=False,
        )

    def to_transitions(self) -> Transitions:
        """Returns a `Transitions` object from the trajectories"""
        states = self.states[:-1][self.actions != -1]
        next_states = self.states[1:][self.actions != -1]
        actions = self.actions[self.actions != -1]
        is_done = (
            next_states.is_sink_state
            if not self.is_backward
            else next_states.is_initial_state
        )
        if self._log_rewards is None:
            log_rewards = None
        else:
            log_rewards = torch.full_like(actions, fill_value=-1.0, dtype=torch.float)
            log_rewards[is_done] = torch.cat(
                [
                    self._log_rewards[self.when_is_done == i]
                    for i in range(self.when_is_done.max() + 1)
                ],
                dim=0,
            )
        log_probs = self.log_probs[self.actions != -1]
        return Transitions(
            env=self.env,
            states=states,
            actions=actions,
            is_done=is_done,
            next_states=next_states,
            is_backward=self.is_backward,
            log_rewards=log_rewards,
            log_probs=log_probs,
        )

    def to_states(self) -> States:
        """Returns a `States` object from the trajectories, containing all states in the trajectories"""
        states = self.states.flatten()
        return states[~states.is_sink_state]

    def to_non_initial_intermediary_and_terminating_states(
        self,
    ) -> tuple[States, States]:
        """Returns a tuple of `States` objects from the trajectories, containing all non-initial intermediary and all terminating states in the trajectories

        Returns:
            Tuple[States, States]: - All the intermediary states in the trajectories that are not s0.
                                   - All the terminating states in the trajectories that are not s0.
        """
        states = self.states
        intermediary_states = states[~states.is_sink_state & ~states.is_initial_state]
        terminating_states = self.last_states
        terminating_states.log_rewards = self.log_rewards
        return intermediary_states, terminating_states



from abc import ABC, abstractmethod
from collections import Counter
from typing import Optional

import torch
from torchtyping import TensorType

from gfn.containers.states import States
from gfn.containers.trajectories import Trajectories
from gfn.envs.env import Env

# Typing
TensorPmf = TensorType["n_states", float]


class TrajectoryDistribution(ABC):
    """
    Represents an abstract distribution over trajectories.
    """

    @abstractmethod
    def sample(self, n_trajectories: int) -> Trajectories:
        """
        Sample a batch of trajectories.
        """
        pass


class TerminatingStatesDistribution(ABC):
    """
    Represents an abstract distribution over terminating states.
    """

    @abstractmethod
    def pmf(self) -> TensorPmf:
        """
        Compute the probability mass function of the distribution.
        """
        pass


class EmpiricalTrajectoryDistribution(TrajectoryDistribution):
    """
    Represents an empirical distribution over trajectories.
    """

    def __init__(self, trajectories: Trajectories):
        self.trajectories = trajectories

    def sample(self, n_trajectories: Optional[int] = None) -> Trajectories:
        if n_trajectories is None:
            return self.trajectories
        return self.trajectories.sample(n_trajectories)


class EmpiricalTerminatingStatesDistribution(TerminatingStatesDistribution):
    """
    Represents an empirical distribution over terminating states.
    """

    def __init__(self, env: Env, states: States) -> None:
        assert len(states.batch_shape) == 1, "States should be a linear batch of states"
        self.states = states
        self.n_states = states.batch_shape[0]
        self.states_to_indices = env.get_terminating_states_indices
        self.env_n_terminating_states = env.n_terminating_states

    def pmf(self) -> TensorPmf:
        states_indices = self.states_to_indices(self.states).cpu().numpy().tolist()
        counter = Counter(states_indices)
        counter_list = [
            counter[state_idx] if state_idx in counter else 0
            for state_idx in range(self.env_n_terminating_states)
        ]
        return torch.tensor(counter_list, dtype=torch.float) / len(states_indices)


class TrajectoryBasedTerminatingStateDistribution(TerminatingStatesDistribution):
    """
    Represents a distribution over final states.
    """

    def __init__(
        self, trajectory_distribution: EmpiricalTrajectoryDistribution
    ) -> None:
        self.trajectory_distribution = trajectory_distribution
        self.states_to_indices = (
            self.trajectory_distribution.trajectories.env.get_terminating_states_indices
        )
        self.env_n_terminating_states = (
            self.trajectory_distribution.trajectories.env.n_terminating_states
        )

    def sample(self, n_final_states: Optional[int] = None) -> States:
        """
        Sample a batch of final states.
        """
        trajectories = self.trajectory_distribution.sample(n_final_states)
        return trajectories.last_states

    def pmf(self) -> TensorPmf:
        """
        Compute the probability mass function of the distribution.
        """
        samples = self.sample()
        samples_indices = self.states_to_indices(samples).cpu().numpy().tolist()
        counter = Counter(samples_indices)
        counter_list = [
            counter[state_idx] if state_idx in counter else 0
            for state_idx in range(self.env_n_terminating_states)
        ]
        return torch.tensor(counter_list, dtype=torch.float) / len(samples_indices)
